import os
from pathlib import Path

import boto3
import pytest

import runhouse as rh
from runhouse.globals import rns_client

from .test_helpers.lambda_tests.basic_handler_layer import arr_handler

from .test_helpers.lambda_tests.basic_test_handler import lambda_no_args, lambda_sum

CUR_WORK_DIR = os.path.dirname(os.path.abspath(__file__))
TEST_RESOURCES = f"{CUR_WORK_DIR}/test_helpers/lambda_tests"

CRED_PATH = f"{Path.home()}/.aws/credentials"
DEFAULT_REGION = "us-east-1"


@pytest.fixture(scope="session")
def basic_function():
    my_lambda = rh.aws_lambda_fn(
        fn=lambda_sum,
        runtime="python3.9",
        name="test_lambda_create_and_run",
    )
    my_lambda.save()
    try:
        yield my_lambda
    finally:
        my_lambda.teardown()
        if rns_client.exists(my_lambda.rns_address):
            rns_client.delete_configs(my_lambda)


@pytest.fixture(scope="session")
def numpy_function():
    name = "test_lambda_numpy"
    my_lambda = rh.aws_lambda_fn(
        fn=arr_handler,
        runtime="python3.9",
        name=name,
        env={"reqs": ["numpy", "pandas"], "env_vars": None},
    ).save()
    yield my_lambda
    my_lambda.teardown()
    if rns_client.exists(my_lambda.rns_address):
        rns_client.delete_configs(my_lambda)


def test_create_and_run_no_layers(basic_function):
    res = basic_function(3, 4)
    assert res == 7
    reload_func = rh.aws_lambda_fn(name=basic_function.name)
    res2 = reload_func(12, 7)
    assert res2 == 19
    reload_func_rns_name = rh.aws_lambda_fn(name=basic_function.rns_address)
    res3 = reload_func_rns_name(4, 6)
    assert res3 == 10


def test_load_not_existing_lambda():
    name_no_user = "test_lambda_create_and_run1"
    with pytest.raises(ValueError) as valueError:
        rh.aws_lambda_fn(name=name_no_user)
    assert str(valueError.value) == f"Could not find a Lambda called {name_no_user}."

    name_with_user = "/sashab/test_lambda_no_such_lambda"
    with pytest.raises(ValueError) as valueError:
        rh.aws_lambda_fn(name=name_with_user)
    assert str(valueError.value) == (
        f"Could not find a Lambda called {name_with_user}."
    )


def test_crate_no_arguments():
    with pytest.raises(RuntimeError) as no_args:
        rh.aws_lambda_fn()
    assert str(no_args.value) == (
        "Please provide a callable function OR use from_handler_file method"
        + "in order to create a Lambda function."
    )


def test_bad_runtime_to_factory():
    name = "test_wrong_runtime"

    wrong_runtime_1 = rh.aws_lambda_fn(fn=lambda_sum, runtime=None, name=f"{name}_1")

    wrong_runtime_2 = rh.aws_lambda_fn(fn=lambda_sum, name=f"{name}_2")

    wrong_runtime_3 = rh.aws_lambda_fn(
        fn=lambda_sum, runtime="python3.91", name=f"{name}_3"
    )

    assert wrong_runtime_1.runtime == "python3.9"
    assert wrong_runtime_2.runtime == "python3.9"
    assert wrong_runtime_3.runtime == "python3.9"

    assert wrong_runtime_1.teardown() is True
    if rns_client.exists(wrong_runtime_1.rns_address):
        rns_client.delete_configs(wrong_runtime_1)

    assert wrong_runtime_2.teardown() is True
    if rns_client.exists(wrong_runtime_2.rns_address):
        rns_client.delete_configs(wrong_runtime_1)

    assert wrong_runtime_3.teardown() is True
    if rns_client.exists(wrong_runtime_3.rns_address):
        rns_client.delete_configs(wrong_runtime_3)


def test_func_no_args():
    name = "test_lambda_no_args"
    my_lambda = rh.aws_lambda_fn(
        fn=lambda_no_args,
        runtime="python3.9",
        name=name,
    )
    assert my_lambda() == "no args lambda"
    assert my_lambda.name == "test_lambda_no_args"
    assert my_lambda.teardown() is True


def test_create_and_run_generate_name():
    my_lambda = rh.aws_lambda_fn(fn=lambda_sum, runtime="python3.9")
    res = my_lambda(3, 4)
    assert res == 7
    my_lambda.save()
    reload_func = rh.aws_lambda_fn(name="lambda_sum")
    res2 = reload_func(12, 7)
    assert res2 == 19
    assert my_lambda.teardown() is True
    assert reload_func.teardown() is True


def test_create_and_run_layers_dict(numpy_function):
    res = numpy_function([1, 2, 3], [1, 2, 3])
    assert res == 12


def test_reload_func_with_libs(numpy_function):
    # tests that after the libs are installed, they are not being re-installed.
    my_reloaded_lambda = rh.aws_lambda_fn(name=numpy_function.name)
    res = my_reloaded_lambda([1, 2, 3], [12, 5, 9])
    assert res == 32


def test_create_and_run_layers_env():
    name = "test_lambda_numpy_env"
    my_env = rh.env(reqs=["numpy"])
    my_lambda = rh.aws_lambda_fn(
        fn=arr_handler,
        runtime="python3.9",
        name=name,
        env=my_env,
    )
    res = my_lambda([1, 2, 3], [2, 5, 6])
    assert res == 19
    assert my_lambda.teardown() is True


def test_create_and_run_layers_list():
    name = "test_lambda_numpy_list"
    my_lambda = rh.aws_lambda_fn(
        fn=arr_handler,
        runtime="python3.9",
        name=name,
        env=["numpy"],
    )
    res = my_lambda([1, 2, 3], [4, 7, 9])
    assert res == 26
    assert my_lambda.teardown() is True


def test_layers_increase_timeout_and_memory():
    name = "test_lambda_numpy_increase_params"
    my_lambda = rh.aws_lambda_fn(
        fn=arr_handler,
        runtime="python3.9",
        name=name,
        env=["numpy"],
        timeout=3,
        memory_size=128,
    )
    res = my_lambda([1, 2, 3], [4, 7, 9])
    assert res == 26
    lambda_client = boto3.client("lambda")
    lambda_config = lambda_client.get_function(FunctionName=my_lambda.name)
    assert lambda_config["Configuration"]["Timeout"] == 600
    assert lambda_config["Configuration"]["MemorySize"] == 10240
    assert lambda_config["Configuration"]["EphemeralStorage"]["Size"] == 10240
    assert lambda_config["Configuration"]["FunctionName"] == my_lambda.name
    assert my_lambda.teardown() is True


def test_create_and_run_layers_txt():
    name = "test_lambda_numpy_txt"
    my_lambda = rh.aws_lambda_fn(
        fn=arr_handler,
        runtime="python3.9",
        name=name,
        env=[f"reqs:{TEST_RESOURCES}/"],
    )
    res = my_lambda([1, 2, 3], [1, 2, 3])
    assert res == 12
    assert my_lambda.teardown() is True


def test_update_lambda_one_file():
    name = "test_lambda_create_and_run"
    my_lambda = rh.aws_lambda_fn(fn=lambda_sum, runtime="python3.9", name=name)
    res = my_lambda(6, 4)
    assert res == 10
    reload_func = rh.aws_lambda_fn(name=name)
    res2 = reload_func(12, 13)
    assert res2 == 25


def test_mult_files_each():
    """The handler function calls function from each file separately.
    For example, there are a.py, b.py and c.py. Each file has the following funcs, respectively: func_a, func_b and
    func_c. So in the test, the main function (handler) will look something like this:
    import a.py, b.py and c.py
    def handler_func:
        func_a()
        func_b()
        func_c()
    Also tests the From_handler_file method.
    """
    prefix = "call_files_separately"
    folder_path = f"{TEST_RESOURCES}/{prefix}"
    handler_paths = os.listdir(folder_path)
    handler_paths = [p for p in handler_paths if ".py" in p]
    handler_paths.sort()
    handler_paths = [f"{folder_path}/{p}" for p in handler_paths]
    name = "test_lambda_multiple_files_s"
    my_lambda_calc_1 = rh.LambdaFunction.from_handler_file(
        paths_to_code=handler_paths,
        handler_function_name="my_calc",
        runtime="python3.9",
        name=name,
        env={"reqs": ["numpy"], "env_vars": None},
    )
    res1 = my_lambda_calc_1(2, 3)
    res2 = my_lambda_calc_1(5, 3)
    res3 = my_lambda_calc_1(2, 7)
    res4 = my_lambda_calc_1(10, 5)
    assert res1 == 2.5
    assert res2 == 3.2
    assert res3 == 22.5
    assert res4 == 7.5
    assert my_lambda_calc_1.teardown() is True


def test_few_python_files_chain():
    """The handler function calls functions from different files in chain.
    For example, there are a.py, b.py and c.py. Each file has the following funcs, respectively: func_a, func_b and
    func_c. So in the test, the main function (handler) will look something like this:
    import c.py
    def handler_func:
        func_c()
    where func_c() calls func_b() which calls func_a().
    Also tests the From_handler_file method.
    """
    prefix = "call_files_chain"
    folder_path = f"{TEST_RESOURCES}/{prefix}"
    handler_paths = os.listdir(folder_path)
    handler_paths = [p for p in handler_paths if ".py" in p]
    handler_paths.sort()
    handler_paths = [f"{folder_path}/{p}" for p in handler_paths]
    name = "test_lambda_multiple_files_c"
    my_lambda_calc_2 = rh.LambdaFunction.from_handler_file(
        paths_to_code=handler_paths,
        handler_function_name="special_calc",
        runtime="python3.9",
        name=name,
    )
    res1 = my_lambda_calc_2(2, 3)
    res2 = my_lambda_calc_2(5, 3)
    res3 = my_lambda_calc_2(2, 7)
    res4 = my_lambda_calc_2(10, 5)
    assert res1 == 16
    assert res2 == 17
    assert res3 == 20
    assert res4 == 20
    assert my_lambda_calc_2.teardown() is True


def test_args(basic_function):
    res1 = basic_function(2, 3)
    res2 = basic_function(5, arg2=3)
    res3 = basic_function(arg1=2, arg2=7)
    assert res1 == 5
    assert res2 == 8
    assert res3 == 9


def test_map_starmap(basic_function):
    res_map1 = basic_function.map([1, 2, 3], [4, 5, 6])
    res_map2 = basic_function.map([6, 2, 3], [15, 52, 61])
    res_map3 = basic_function.starmap([(1, 2), (3, 4), (5, 6)])
    res_map4 = basic_function.starmap([(12, 5), (44, 32), (8, 3)])
    assert res_map1 == [5, 7, 9]
    assert res_map2 == [21, 54, 64]
    assert res_map3 == [3, 7, 11]
    assert res_map4 == [17, 76, 11]


def test_create_from_config():
    handler_path = [f"{TEST_RESOURCES}/basic_test_handler.py"]
    name = "test_lambda_from_config"
    config = {
        "paths_to_code": handler_path,
        "handler_function_name": "lambda_sum",
        "runtime": "python3.9",
        "args_names": ["arg1", "arg2"],
        "name": name,
        "tmp_size": 3072,
    }
    config_lambda = rh.LambdaFunction.from_config(config)
    res1 = config_lambda(1, 2)
    res2 = config_lambda(8, 12)
    res3 = config_lambda(14, 17)

    assert res1 == 3
    assert res2 == 20
    assert res3 == 31
    assert config_lambda.teardown() is True


def test_delete_lambda():
    lambda_client = boto3.client("lambda")
    iam_client = boto3.client("iam")
    logs_client = boto3.client("logs")
    name = "test_lambda_to_delete"
    lambda_to_delete = rh.aws_lambda_fn(fn=lambda_sum, runtime="python3.9", name=name)

    lambda_to_delete.save()
    assert lambda_to_delete(5, 11) == 16

    lambda_name = lambda_to_delete.name
    lambda_policy = f"{lambda_name}_Policy"
    lambda_role = f"{lambda_name}_Role"
    lambda_log_group = f"/aws/lambda/{lambda_name}"

    del_res = lambda_to_delete.teardown()
    assert del_res is True

    functions_in_aws = [
        function["FunctionName"]
        for function in lambda_client.list_functions()["Functions"]
    ]
    assert lambda_name not in functions_in_aws

    customer_managed_policies = [
        policy["PolicyName"]
        for policy in iam_client.list_policies(Scope="Local")["Policies"]
    ]
    assert lambda_policy not in customer_managed_policies

    roles_in_aws = [role["RoleName"] for role in iam_client.list_roles()["Roles"]]
    assert lambda_role not in roles_in_aws

    logs_groups_in_aws = [
        log_group["logGroupName"]
        for log_group in logs_client.describe_log_groups()["logGroups"]
    ]
    assert lambda_log_group not in logs_groups_in_aws

    if rns_client.exists(lambda_to_delete.rns_address):
        rns_client.delete_configs(lambda_to_delete)



import logging
import subprocess

from pathlib import Path
from typing import Dict, List, Optional

import yaml

from runhouse.constants import CONDA_INSTALL_CMDS, EMPTY_DEFAULT_ENV_NAME
from runhouse.globals import rns_client
from runhouse.resources.resource import Resource
from runhouse.utils import locate_working_dir, run_with_logs


def _process_reqs(reqs):
    preprocessed_reqs = []
    for package in reqs:
        from runhouse.resources.packages import Package

        # TODO [DG] the following is wrong. RNS address doesn't have to start with '/'. However if we check if each
        #  string exists in RNS this will be incredibly slow, so leave it for now.
        if isinstance(package, str):
            if package[0] == "/" and rns_client.exists(package):
                # If package is an rns address
                package = rns_client.load_config(package)
            else:
                # if package refers to a local path package
                path = Path(Package.split_req_install_method(package)[1]).expanduser()
                if path.is_absolute() or (locate_working_dir() / path).exists():
                    package = Package.from_string(package)
        elif isinstance(package, dict):
            package = Package.from_config(package)
        preprocessed_reqs.append(package)
    return preprocessed_reqs


def _process_env_vars(env_vars):
    processed_vars = (
        _env_vars_from_file(env_vars) if isinstance(env_vars, str) else env_vars
    )
    return processed_vars


def _get_env_from(env, load: bool = True):
    if isinstance(env, Resource):
        return env

    from runhouse.resources.envs import Env

    if isinstance(env, List):
        if len(env) == 0:
            return Env(reqs=env, working_dir=None)
        return Env(reqs=env)
    elif isinstance(env, Dict):
        return Env.from_config(env)
    elif isinstance(env, str) and EMPTY_DEFAULT_ENV_NAME not in env:
        if not load:
            return env

        try:
            return (
                Env.from_name(env)
                if rns_client.exists(env, resource_type="env")
                else env
            )
        except ValueError:
            return env
    return env


def _get_conda_yaml(conda_env=None):
    if not conda_env:
        return None
    if isinstance(conda_env, str):
        if Path(conda_env).expanduser().exists():  # local yaml path
            conda_yaml = yaml.safe_load(open(conda_env))
        elif f"\n{conda_env} " in subprocess.check_output(
            "conda info --envs".split(" ")
        ).decode("utf-8"):
            res = subprocess.check_output(
                f"conda env export -n {conda_env} --no-build".split(" ")
            ).decode("utf-8")
            conda_yaml = yaml.safe_load(res)
        else:
            raise Exception(
                f"{conda_env} must be a Dict or point to an existing path or conda environment."
            )
    else:
        conda_yaml = conda_env

    # ensure correct version to Ray -- this is subject to change if SkyPilot adds additional ray version support
    conda_yaml["dependencies"] = (
        conda_yaml["dependencies"] if "dependencies" in conda_yaml else []
    )
    if not [dep for dep in conda_yaml["dependencies"] if "pip" in dep]:
        conda_yaml["dependencies"].append("pip")
    if not [
        dep
        for dep in conda_yaml["dependencies"]
        if isinstance(dep, Dict) and "pip" in dep
    ]:
        conda_yaml["dependencies"].append({"pip": ["ray >= 2.2.0, != 2.6.0"]})
    else:
        for dep in conda_yaml["dependencies"]:
            if (
                isinstance(dep, Dict)
                and "pip" in dep
                and not [pip for pip in dep["pip"] if "ray" in pip]
            ):
                dep["pip"].append("ray >= 2.2.0, != 2.6.0")
                continue
    return conda_yaml


def _env_vars_from_file(env_file):
    try:
        from dotenv import dotenv_values, find_dotenv
    except ImportError:
        raise ImportError(
            "`dotenv` package is needed. You can install it with `pip install python-dotenv`."
        )

    dotenv_path = find_dotenv(str(env_file), usecwd=True)
    env_vars = dotenv_values(dotenv_path)
    return dict(env_vars)


# ------- Installation helpers -------
def run_setup_command(
    cmd: str,
    cluster: "Cluster" = None,
    env_vars: Dict = None,
    stream_logs: bool = True,
    node: Optional[str] = None,
):
    """
    Helper function to run a command during possibly the cluster default env setup. If a cluster is provided,
    run command on the cluster using SSH. If the cluster is not provided, run locally, as if already on the
    cluster (rpc call).

    Args:
        cmd (str): Command to run on the
        cluster (Optional[Cluster]): (default: None)
        stream_logs (bool): (default: True)

    Returns:
       (status code, stdout)
    """
    if not cluster:
        return run_with_logs(cmd, stream_logs=stream_logs, require_outputs=True)[:2]
    elif cluster.on_this_cluster():
        cmd = cluster.default_env._full_command(cmd)
        return run_with_logs(cmd, stream_logs=stream_logs, require_outputs=True)[:2]

    return cluster._run_commands_with_runner(
        [cmd], stream_logs=stream_logs, env_vars=env_vars, node=node
    )[0]


def install_conda(cluster: "Cluster" = None, node: Optional[str] = None):
    if run_setup_command("conda --version", cluster=cluster, node=node)[0] != 0:
        logging.info("Conda is not installed. Installing...")
        for cmd in CONDA_INSTALL_CMDS:
            run_setup_command(cmd, cluster=cluster, node=node, stream_logs=True)
        if run_setup_command("conda --version", cluster=cluster, node=node)[0] != 0:
            raise RuntimeError("Could not install Conda.")



import copy
import re
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Union

from runhouse.logger import get_logger

from runhouse.resources.envs.utils import install_conda, run_setup_command
from runhouse.resources.hardware.cluster import Cluster
from runhouse.resources.hardware.utils import (
    _get_cluster_from,
    detect_cuda_version_or_cpu,
)
from runhouse.resources.resource import Resource
from runhouse.utils import (
    find_locally_installed_version,
    get_local_install_path,
    is_python_package_string,
    locate_working_dir,
)


INSTALL_METHODS = {"local", "reqs", "pip", "conda", "rh"}

logger = get_logger(__name__)


class CodeSyncError(Exception):
    pass


@dataclass
class InstallTarget:
    local_path: str
    _path_to_sync_to_on_cluster: Optional[str] = None

    @property
    def path_to_sync_to_on_cluster(self) -> str:
        return (
            self._path_to_sync_to_on_cluster
            if self._path_to_sync_to_on_cluster
            else f"~/{Path(self.full_local_path_str()).name}"
        )

    def full_local_path_str(self) -> str:
        return str(Path(self.local_path).expanduser().resolve())

    def __str__(self):
        return f"InstallTarget(local_path={self.local_path}, path_to_sync_to_on_cluster={self._path_to_sync_to_on_cluster})"


class Package(Resource):
    RESOURCE_TYPE = "package"

    # https://pytorch.org/get-started/locally/
    # Note: no binaries exist for 11.4 (https://github.com/pytorch/pytorch/issues/75992)
    TORCH_INDEX_URLS = {
        "11.3": "https://download.pytorch.org/whl/cu113",
        "11.5": "https://download.pytorch.org/whl/cu115",
        "11.6": "https://download.pytorch.org/whl/cu116",
        "11.7": "https://download.pytorch.org/whl/cu117",
        "11.8": "https://download.pytorch.org/whl/cu118",
        "12.1": "https://download.pytorch.org/whl/cu121",
        "cpu": "https://download.pytorch.org/whl/cpu",
    }

    def __init__(
        self,
        name: Optional[str] = None,
        install_method: Optional[str] = None,
        install_target: Optional[Union[str, "Folder"]] = None,
        install_args: Optional[str] = None,
        preferred_version: Optional[str] = None,
        dryrun: bool = False,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        Runhouse Package resource.

        .. note::
            To create a git package, please use the factory method :func:`package`.
        """
        super().__init__(
            name=name,
            dryrun=dryrun,
        )
        self.install_method = install_method
        self.install_target = install_target
        self.install_args = install_args
        self.preferred_version = preferred_version

    def config(self, condensed: bool = True):
        # If the package is just a simple Package.from_string string, no
        # need to store it in rns, just give back the string.
        # if self.install_method in ['pip', 'conda', 'git']:
        #     return f'{self.install_method}:{self.name}'
        config = super().config(condensed)
        config["install_method"] = self.install_method
        config["install_target"] = (
            (
                self.install_target.local_path,
                self.install_target._path_to_sync_to_on_cluster,
            )
            if isinstance(self.install_target, InstallTarget)
            else self.install_target
        )
        config["install_args"] = self.install_args
        config["preferred_version"] = self.preferred_version
        return config

    def __str__(self):
        if self.name:
            return f"Package: {self.name}"
        return f"Package: {self.install_target}"

    @staticmethod
    def _prepend_python_executable(
        install_cmd: str, env: Union[str, "Env"] = None, cluster: "Cluster" = None
    ):
        return (
            f"python3 -m {install_cmd}"
            if cluster or env
            else f"{sys.executable} -m {install_cmd}"
        )

    @staticmethod
    def _prepend_env_command(install_cmd: str, env: Union[str, "Env"] = None):
        if env:
            from runhouse.resources.envs.utils import _get_env_from

            env = _get_env_from(env)
            install_cmd = env._full_command(install_cmd)

        return install_cmd

    def _validate_folder_path(self):
        # If self.path is the same as the user's home directory, raise an error.
        # Check this with Path and expanduser to handle both relative and absolute paths.
        if isinstance(
            self.install_target, InstallTarget
        ) and self.install_target.full_local_path_str() in [
            str(Path("~").expanduser()),
            str(Path("/")),
        ]:
            raise CodeSyncError(
                "Cannot sync the home directory. Please include a Python configuration file in a subdirectory."
            )

    def _pip_install_cmd(
        self, env: Union[str, "Env"] = None, cluster: "Cluster" = None
    ):
        install_args = f" {self.install_args}" if self.install_args else ""
        if isinstance(self.install_target, InstallTarget):
            install_cmd = self.install_target.full_local_path_str() + install_args
        else:
            install_target = f'"{self.install_target}"'
            install_cmd = install_target + install_args

        install_cmd = f"pip install {self._install_cmd_for_torch(install_cmd, cluster)}"
        install_cmd = self._prepend_python_executable(
            install_cmd, cluster=cluster, env=env
        )
        install_cmd = self._prepend_env_command(install_cmd, env=env)
        return install_cmd

    def _conda_install_cmd(
        self, env: Union[str, "Env"] = None, cluster: "Cluster" = None
    ):
        install_args = f" {self.install_args}" if self.install_args else ""
        if isinstance(self.install_target, InstallTarget):
            install_cmd = f"{self.install_target.local_path}" + install_args
        else:
            install_cmd = self.install_target + install_args

        install_cmd = f"conda install -y {install_cmd}"
        install_cmd = self._prepend_env_command(install_cmd, env=env)
        install_conda(cluster=cluster)
        return install_cmd

    def _reqs_install_cmd(
        self, env: Union[str, "Env"] = None, cluster: "Cluster" = None
    ):
        install_args = f" {self.install_args}" if self.install_args else ""
        if not isinstance(self.install_target, InstallTarget):
            install_cmd = self.install_target + install_args
        else:
            on_cluster_path = self.install_target.local_path

            # If not cluster, we're on the cluster, and we must deal with the path locally
            if not cluster:
                reqs_path = f"{on_cluster_path}/requirements.txt"
                if not Path(reqs_path).expanduser().exists():
                    return None

                with open(str(Path(reqs_path).expanduser())) as f:
                    reqs_list = f.readlines()

            # Otherwise, make sure the target folder is on the cluster,
            # and read reqs from the cluster
            else:
                if "requirements.txt" not in cluster._folder_ls(
                    path=on_cluster_path, full_paths=False
                ):
                    return None

                reqs_list = (
                    cluster._folder_get(f"{on_cluster_path}/requirements.txt", mode="r")
                    .strip("\n")
                    .split("\n")
                )
                reqs_path = f"{on_cluster_path}/requirements.txt"

            install_cmd = self._reqs_install_cmd_for_torch(
                reqs_path, reqs_list, install_args, cluster=cluster
            )

        install_cmd = f"pip install {install_cmd}"
        install_cmd = self._prepend_python_executable(
            install_cmd, env=env, cluster=cluster
        )
        install_cmd = self._prepend_env_command(install_cmd, env=env)
        return install_cmd

    def _install(
        self,
        env: Union[str, "Env"] = None,
        cluster: "Cluster" = None,
        node: Optional[str] = None,
    ):
        """Install package.

        Args:
            env (Env or str): Environment to install package on. If left empty, defaults to base environment.
                (Default: ``None``)
            cluster (Optional[Cluster]): If provided, will install package on cluster using SSH.
        """
        logger.info(f"Installing {str(self)} with method {self.install_method}.")

        if isinstance(self.install_target, InstallTarget):
            if cluster and Path(self.install_target.local_path).expanduser().exists():
                cluster.rsync(
                    source=str(self.install_target.local_path),
                    dest=str(self.install_target.path_to_sync_to_on_cluster),
                    up=True,
                    contents=True,
                    node=node,
                )

                self.install_target.local_path = (
                    self.install_target.path_to_sync_to_on_cluster
                )

            path = self.install_target.local_path

            if not path:
                return

        if self.install_method == "pip":

            # If this is a generic pip package, with no version pinned, we want to check if there is a version
            # already installed. If there is, then we ignore preferred version and leave the existing version.
            # The user can always force a version install by doing `numpy==2.0.0` for example. Else, we install
            # the preferred version, that matches their local.
            if (
                is_python_package_string(self.install_target)
                and self.preferred_version is not None
            ):
                # Check if this is installed
                retcode = run_setup_command(
                    f"python -c \"import importlib.util; exit(0) if importlib.util.find_spec('{self.install_target}') else exit(1)\"",
                    cluster=cluster,
                    node=node,
                )[0]
                if retcode != 0:
                    self.install_target = (
                        f"{self.install_target}=={self.preferred_version}"
                    )

            install_cmd = self._pip_install_cmd(env=env, cluster=cluster)
            logger.info(f"Running via install_method pip: {install_cmd}")
            retcode = run_setup_command(install_cmd, cluster=cluster, node=node)[0]
            if retcode != 0:
                raise RuntimeError(
                    f"Pip install {install_cmd} failed, check that the package exists and is available for your platform."
                )

        elif self.install_method == "conda":
            install_cmd = self._conda_install_cmd(env=env, cluster=cluster)
            logger.info(f"Running via install_method conda: {install_cmd}")
            retcode = run_setup_command(install_cmd, cluster=cluster, node=node)[0]
            if retcode != 0:
                raise RuntimeError(
                    f"Conda install {install_cmd} failed, check that the package exists and is "
                    "available for your platform."
                )

        elif self.install_method == "reqs":
            install_cmd = self._reqs_install_cmd(env=env, cluster=cluster)
            if install_cmd:
                logger.info(f"Running via install_method reqs: {install_cmd}")
                retcode = run_setup_command(install_cmd, cluster=cluster, node=node)[0]
                if retcode != 0:
                    raise RuntimeError(
                        f"Reqs install {install_cmd} failed, check that the package exists and is available for your platform."
                    )
            else:
                logger.info(f"{path}/requirements.txt not found, skipping reqs install")

        else:
            if self.install_method != "local":
                raise ValueError(
                    f"Unknown install method {self.install_method}. Must be one of {INSTALL_METHODS}"
                )

        # Need to append to path
        if self.install_method in ["local", "reqs"]:
            if isinstance(self.install_target, InstallTarget):
                sys.path.insert(
                    0, self.install_target.full_local_path_str()
                ) if not cluster else run_setup_command(
                    f"export PATH=$PATH;{self.install_target.full_local_path_str()}",
                    cluster=cluster,
                    node=node,
                )
            elif not cluster:
                if Path(self.install_target).resolve().expanduser().exists():
                    sys.path.insert(
                        0, str(Path(self.install_target).resolve().expanduser())
                    )
                else:
                    raise ValueError(
                        f"install_target {self.install_target} must be a Folder or a path to a directory for install_method {self.install_method}"
                    )
            else:
                raise ValueError(
                    f"If cluster is provided, install_target must be a Folder for install_method {self.install_method}"
                )

    # ----------------------------------
    # Torch Install Helpers
    # ----------------------------------
    def _reqs_install_cmd_for_torch(
        self, reqs_path, reqs_list, install_args="", cluster=None
    ):
        """Read requirements from file, append --index-url and --extra-index-url where relevant for torch packages,
        and return list of formatted packages."""
        # if torch extra index url is already defined by the user or torch isn't a req, directly pip install reqs file
        if not any("torch" in req for req in reqs_list):
            return f"-r {reqs_path}" + install_args

        cuda_version_or_cpu = detect_cuda_version_or_cpu(cluster=cluster)
        for req in reqs_list:
            if (
                "--index-url" in req or "--extra-index-url" in req
            ) and "pytorch.org" in req:
                return f"-r {reqs_path}" + install_args

        # add extra-index-url for torch if not found
        return f"-r {reqs_path} --extra-index-url {self._torch_index_url(cuda_version_or_cpu)}"

    def _install_cmd_for_torch(self, install_cmd, cluster=None):
        """Return the correct formatted pip install command for the torch package(s) provided."""
        if install_cmd.startswith("#"):
            return None

        torch_source_packages = ["torch", "torchvision", "torchaudio"]
        if not any([x in install_cmd for x in torch_source_packages]):
            return install_cmd

        packages_to_install: list = self._packages_to_install_from_cmd(install_cmd)
        final_install_cmd = ""
        cuda_version_or_cpu = detect_cuda_version_or_cpu(cluster=cluster)
        for package_install_cmd in packages_to_install:
            formatted_cmd = self._install_url_for_torch_package(
                package_install_cmd, cuda_version_or_cpu
            )
            if formatted_cmd:
                final_install_cmd += formatted_cmd + " "

        final_install_cmd = final_install_cmd.rstrip()
        return final_install_cmd if final_install_cmd != "" else None

    def _install_url_for_torch_package(self, install_cmd, cuda_version_or_cpu):
        """Build the full install command, adding a --index-url and --extra-index-url where applicable."""
        # Grab the relevant index url for torch based on the CUDA version provided
        if "," in install_cmd:
            # If installing a range of versions format the string to make it compatible with `pip_install` method
            install_cmd = install_cmd.replace(" ", "")

        index_url = self._torch_index_url(cuda_version_or_cpu)
        if index_url and not any(
            specifier in install_cmd for specifier in ["--index-url ", "-i "]
        ):
            install_cmd = f"{install_cmd} --index-url {index_url}"

        if "--extra-index-url" not in install_cmd:
            return f"{install_cmd} --extra-index-url https://pypi.python.org/simple/"

        return install_cmd

    def _torch_index_url(self, cuda_version_or_cpu: str):
        return self.TORCH_INDEX_URLS.get(cuda_version_or_cpu)

    @staticmethod
    def _packages_to_install_from_cmd(install_cmd: str):
        """Split a string of command(s) into a list of separate commands"""
        # Remove any --extra-index-url flags from the install command (to be added later by default)
        install_cmd = re.sub(r"--extra-index-url\s+\S+", "", install_cmd)
        install_cmd = install_cmd.strip()

        if ", " in install_cmd:
            # Ex: 'torch>=1.13.0,<2.0.0'
            return [install_cmd]

        matches = re.findall(r"(\S+(?:\s+(-i|--index-url)\s+\S+)?)", install_cmd)

        packages_to_install = [match[0] for match in matches]
        return packages_to_install

    def to(
        self,
        system: Union[str, Dict, "Cluster"],
        path: Optional[str] = None,
    ):
        """Copy the package onto filesystem or cluster, and return the new Package object.

        Args:
            system (str, Dict, or Cluster): Cluster to send the package to.
        """
        if not isinstance(self.install_target, InstallTarget):
            raise TypeError(
                "`install_target` must be an InstallTarget in order to copy the package to a system."
            )

        system = _get_cluster_from(system)
        if isinstance(system, Cluster) and system.on_this_cluster():
            return self

        self._validate_folder_path()
        if isinstance(system, Cluster):
            system.rsync(
                source=str(self.install_target.full_local_path_str()),
                dest=str(self.install_target.path_to_sync_to_on_cluster),
                up=True,
                contents=True,
            )

            new_package = copy.copy(self)
            new_package.install_target = InstallTarget(
                local_path=self.install_target.path_to_sync_to_on_cluster,
                _path_to_sync_to_on_cluster=self.install_target.path_to_sync_to_on_cluster,
            )
            return new_package

        return self

    @staticmethod
    def split_req_install_method(req_str: str):
        """Split a requirements string into a install method and the rest of the string."""
        splat = req_str.split(":", 1)
        return (splat[0], splat[1]) if len(splat) > 1 else ("", splat[0])

    @staticmethod
    def from_config(config: Dict, dryrun: bool = False, _resolve_children: bool = True):
        if isinstance(config.get("install_target"), tuple):
            config["install_target"] = InstallTarget(
                local_path=config["install_target"][0],
                _path_to_sync_to_on_cluster=config["install_target"][1],
            )
        if config.get("resource_subtype") == "GitPackage":
            from runhouse import GitPackage

            return GitPackage.from_config(
                config, dryrun=dryrun, _resolve_children=_resolve_children
            )
        return Package(**config, dryrun=dryrun)

    @staticmethod
    def from_string(specifier: str, dryrun: bool = False):
        if specifier == "requirements.txt":
            specifier = "reqs:./"

        # Use regex to check if specifier matches '<method>:https://github.com/<path>' or 'https://github.com/<path>'
        match = re.search(
            r"^(?:(?P<method>[^:]+):)?(?P<path>https://github.com/.+)", specifier
        )
        if match:
            install_method = match.group("method")
            url = match.group("path")
            from runhouse.resources.packages.git_package import git_package

            return git_package(
                git_url=url, install_method=install_method, dryrun=dryrun
            )

        install_method, target_and_args = Package.split_req_install_method(specifier)

        # Handles a case like "torch --index-url https://download.pytorch.org/whl/cu113"
        rel_target, args = (
            target_and_args.split(" ", 1)
            if " " in target_and_args
            else (target_and_args, "")
        )

        # We need to do this because relative paths are relative to the current working directory!
        abs_target = (
            Path(rel_target).expanduser()
            if Path(rel_target).expanduser().is_absolute()
            else Path(locate_working_dir()) / rel_target
        )
        if abs_target.exists():
            target = InstallTarget(
                local_path=str(abs_target), _path_to_sync_to_on_cluster=None
            )
        else:
            target = rel_target

        # If install method is not provided, we need to infer it
        if not install_method:
            if Path(specifier).resolve().exists():
                install_method = "reqs"
            else:
                install_method = "pip"

        # If we are just defaulting to pip, attempt to install the same version of the package
        # that is already installed locally
        # Check if the target is only letters, nothing else. This means its a string like 'numpy'.
        preferred_version = None
        if install_method == "pip" and is_python_package_string(target):
            locally_installed_version = find_locally_installed_version(target)
            if locally_installed_version:
                # Check if this is a package that was installed from local
                local_install_path = get_local_install_path(target)
                if local_install_path and Path(local_install_path).exists():
                    target = InstallTarget(
                        local_path=local_install_path, _path_to_sync_to_on_cluster=None
                    )

                else:
                    # We want to preferrably install this version of the package server-side
                    preferred_version = locally_installed_version

        # "Local" install method is a special case where we just copy a local folder and add to path
        if install_method == "local":
            return Package(
                install_target=target, install_method=install_method, dryrun=dryrun
            )

        elif install_method in ["reqs", "pip", "conda"]:
            return Package(
                install_target=target,
                install_args=args,
                install_method=install_method,
                preferred_version=preferred_version,
                dryrun=dryrun,
            )
        elif install_method == "rh":
            # Calling the factory method below
            return package(name=specifier[len("rh:") :], dryrun=dryrun)
        else:
            raise ValueError(
                f"Unknown install method {install_method}. Must be one of {INSTALL_METHODS}"
            )


def package(
    name: str = None,
    install_method: str = None,
    install_str: str = None,
    path: str = None,
    system: str = None,
    load_from_den: bool = True,
    dryrun: bool = False,
) -> Package:
    """
    Builds an instance of :class:`Package`.

    Args:
        name (str, optional): Name to assign the package resource.
        install_method (str, optional): Method for installing the package.
            Options: [``pip``, ``conda``, ``reqs``, ``local``]
        install_str (str, optional): Additional arguments to install.
        path (str, optional): URL of the package to install.
        system (str, optional): File system or cluster on which the package lives.
            Currently this must a cluster or one of: [``file``, ``s3``, ``gs``].
        load_from_den (bool, optional): Whether to try loading the Package from Den. (Default: ``True``)
        dryrun (bool, optional): Whether to create the Package if it doesn't exist, or load the Package
            object as a dryrun. (Default: ``False``)

    Returns:
        Package: The resulting package.

    Example:
        >>> import runhouse as rh
        >>> reloaded_package = rh.package(name="my-package")
        >>> local_package = rh.package(path="local/folder/path", install_method="local")
    """
    if name and not any([install_method, install_str, path, system]):
        # If only the name is provided and dryrun is set to True
        return Package.from_name(name, load_from_den=load_from_den, dryrun=dryrun)

    install_target = None
    install_args = None
    if path is not None:
        install_target = (path, None)
        install_args = install_str
    elif install_str is not None:
        install_target, install_args = install_str.split(" ", 1)

    return Package(
        install_method=install_method,
        install_target=install_target,
        install_args=install_args,
        name=name,
        dryrun=dryrun,
    )



import subprocess

from pathlib import Path
from typing import Dict, List, Optional, Union

import yaml

from runhouse.constants import CONDA_PREFERRED_PYTHON_VERSION, ENVS_DIR
from runhouse.globals import obj_store
from runhouse.logger import get_logger

from runhouse.resources.envs.utils import install_conda, run_setup_command
from runhouse.resources.packages import Package

from .env import Env

logger = get_logger(__name__)


class CondaEnv(Env):
    RESOURCE_TYPE = "env"

    def __init__(
        self,
        conda_yaml: Union[str, Dict],
        name: Optional[str] = None,
        reqs: List[Union[str, Package]] = [],
        setup_cmds: List[str] = None,
        env_vars: Optional[Dict] = {},
        working_dir: Optional[Union[str, Path]] = "./",
        secrets: List[Union[str, "Secret"]] = [],
        dryrun: bool = True,
        **kwargs,  # We have this here to ignore extra arguments when calling from_config
    ):
        """
        Runhouse CondaEnv object.

        .. note::
            To create a CondaEnv, please use the factory methods :func:`env` or :func:`conda_env`.
        """
        self.reqs = reqs
        self.conda_yaml = conda_yaml  # dict representing conda env
        super().__init__(
            name=name,
            reqs=reqs,
            setup_cmds=setup_cmds,
            env_vars=env_vars,
            working_dir=working_dir,
            secrets=secrets,
            dryrun=dryrun,
        )

    def config(self, condensed=True):
        config = super().config(condensed)
        config.update({"conda_yaml": self.conda_yaml})
        return config

    @property
    def env_name(self):
        return self.conda_yaml["name"]

    def _create_conda_env(
        self, force: bool = False, cluster: "Cluster" = None, node: Optional[str] = None
    ):
        yaml_path = Path(ENVS_DIR) / f"{self.env_name}.yml"

        env_exists = (
            f"\n{self.env_name} "
            in run_setup_command("conda info --envs", cluster=cluster, node=node)[1]
        )
        run_setup_command(f"mkdir -p {ENVS_DIR}", cluster=cluster, node=node)
        yaml_exists = (
            (Path(ENVS_DIR).expanduser() / f"{self.env_name}.yml").exists()
            if not cluster
            else run_setup_command(f"ls {yaml_path}", cluster=cluster, node=node)[0]
            == 0
        )

        if force or not (yaml_exists and env_exists):
            # dump config into yaml file on cluster
            if not cluster:
                python_commands = "; ".join(
                    [
                        "import yaml",
                        "from pathlib import Path",
                        f"path = Path('{ENVS_DIR}').expanduser()",
                        f"yaml.dump({self.conda_yaml}, open(path / '{self.env_name}.yml', 'w'))",
                    ]
                )
                subprocess.run(f'python -c "{python_commands}"', shell=True)
            else:
                contents = yaml.dump(self.conda_yaml)
                run_setup_command(
                    f"echo $'{contents}' > {yaml_path}", cluster=cluster, node=node
                )

            # create conda env from yaml file
            run_setup_command(
                f"conda env create -f {yaml_path}", cluster=cluster, node=node
            )

            env_exists = (
                f"\n{self.env_name} "
                in run_setup_command("conda info --envs", cluster=cluster, node=node)[1]
            )
            if not env_exists:
                raise RuntimeError(f"conda env {self.env_name} not created properly.")

    def install(
        self, force: bool = False, cluster: "Cluster" = None, node: Optional[str] = None
    ):
        """Locally install packages and run setup commands.

        Args:
            force (bool, optional): Whether to force re-install env if it has already been installed.
                (default: ``False``)
            cluster (bool, optional): If None, installs env locally. Otherwise installs remotely
                on the cluster using SSH. (default: ``None``)
        """
        if not any(["python" in dep for dep in self.conda_yaml["dependencies"]]):
            status_codes = run_setup_command(
                "python --version", cluster=cluster, node=node
            )
            base_python_version = (
                status_codes[1].split()[1]
                if status_codes[0] == 0
                else CONDA_PREFERRED_PYTHON_VERSION
            )
            self.conda_yaml["dependencies"].append(f"python=={base_python_version}")
        install_conda(cluster=cluster)
        local_env_exists = (
            f"\n{self.env_name} "
            in run_setup_command("conda info --envs", cluster=cluster, node=node)[1]
        )

        # If we're doing the install remotely via SSH (e.g. for default_env), there is no cache
        if not cluster:
            # Hash the config_for_rns to check if we need to create/install the conda env
            env_config = self.config()
            # Remove the name because auto-generated names will be different, but the installed components are the same
            env_config.pop("name")
            install_hash = hash(str(env_config))
            # Check the existing hash
            if (
                local_env_exists
                and install_hash in obj_store.installed_envs
                and not force
            ):
                logger.debug("Env already installed, skipping")
                return
            obj_store.installed_envs[install_hash] = self.name

        self._create_conda_env(force=force, cluster=cluster, node=node)

        self._install_reqs(cluster=cluster, node=node)
        self._run_setup_cmds(cluster=cluster, node=node)

        return

    @property
    def _run_cmd(self):
        """Command prefix to run on Conda Env."""
        return f"conda run -n {self.env_name}"

    @property
    def _activate_cmd(self):
        """Command to activate Conda Env."""
        return f"conda activate {self.env_name}"



from .cluster import Cluster
from .cluster_factory import cluster, kubernetes_cluster, ondemand_cluster
from .on_demand_cluster import OnDemandCluster
from .utils import (
    _current_cluster,
    _default_env_if_on_cluster,
    _get_cluster_from,
    cluster_config_file_exists,
    load_cluster_config_from_file,
)



import hashlib
import json
import subprocess

from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Union

from runhouse.constants import (
    CLUSTER_CONFIG_PATH,
    EMPTY_DEFAULT_ENV_NAME,
    RESERVED_SYSTEM_NAMES,
)

from runhouse.logger import get_logger
from runhouse.resources.envs.utils import _get_env_from, run_setup_command
from runhouse.resources.hardware.sky.command_runner import (
    _HASH_MAX_LENGTH,
    ssh_options_list,
    SshMode,
)

logger = get_logger(__name__)


class ServerConnectionType(str, Enum):
    """Manage the type of connection Runhouse will make with the API server started on the cluster.
    ``ssh``: Use port forwarding to connect to the server via SSH, by default on port 32300.
    ``tls``: Do not use port forwarding and start the server with HTTPS (using custom or fresh TLS certs), by default
        on port 443.
    ``none``: Do not use port forwarding, and start the server with HTTP, by default on port 80.
    """

    SSH = "ssh"
    TLS = "tls"
    NONE = "none"


class ResourceServerStatus(str, Enum):
    running = "running"
    terminated = "terminated"
    unauthorized = "unauthorized"
    unknown = "unknown"
    internal_server_error = "internal_server_error"
    runhouse_daemon_down = "runhouse_daemon_down"
    invalid_url = "invalid_url"
    local_cluster = "local_cluster"


def cluster_config_file_exists() -> bool:
    return Path(CLUSTER_CONFIG_PATH).expanduser().exists()


def load_cluster_config_from_file() -> Dict:
    if cluster_config_file_exists():
        with open(Path(CLUSTER_CONFIG_PATH).expanduser()) as f:
            cluster_config = json.load(f)
        return cluster_config
    else:
        return {}


def _current_cluster(key="config"):
    """Retrieve key value from the current cluster config.
    If key is "config", returns entire config."""
    from runhouse.globals import obj_store

    cluster_config = obj_store.get_cluster_config()
    cluster_config.pop("creds", None)
    if cluster_config:
        # This could be a local cluster started via runhouse start,
        # in which case it would have no Name.
        if key in ["cluster_name", "name"] and "name" not in cluster_config:
            return None
        if key == "config":
            return cluster_config
        if key == "cluster_name":
            return cluster_config["name"].rsplit("/", 1)[-1]
        return cluster_config[key]
    else:
        return None


def _default_env_if_on_cluster():
    from runhouse import Env

    config = _current_cluster()
    return (
        _get_env_from(
            config.get(
                "default_env",
                Env(name=EMPTY_DEFAULT_ENV_NAME),
            )
        )
        if config
        else None
    )


def _get_cluster_from(system, dryrun=False):
    from .cluster import Cluster

    if isinstance(system, Cluster):
        return system
    if system in RESERVED_SYSTEM_NAMES:
        return system

    if isinstance(system, Dict):
        return Cluster.from_config(system, dryrun)

    if isinstance(system, str):
        config = _current_cluster(key="config")
        if config and system == config.get("name"):
            return Cluster.from_config(config, dryrun)
        try:
            system = Cluster.from_name(name=system, dryrun=dryrun)
        except ValueError:
            # Name not found in RNS. Doing the lookup this way saves us a hop to RNS
            pass

    return system


def _unnamed_default_env_name(cluster_name):
    return f"{cluster_name}_default_env"


def detect_cuda_version_or_cpu(cluster: "Cluster" = None, node: Optional[str] = None):
    """Return the CUDA version on the cluster. If we are on a CPU-only cluster return 'cpu'.

    Note: A cpu-only machine may have the CUDA toolkit installed, which means nvcc will still return
    a valid version. Also check if the NVIDIA driver is installed to confirm we are on a GPU."""

    status_codes = run_setup_command("nvcc --version", cluster=cluster, node=node)
    if not status_codes[0] == 0:
        return "cpu"
    cuda_version = status_codes[1].split("release ")[1].split(",")[0]

    if run_setup_command("nvidia-smi", cluster=cluster, node=node)[0] == 0:
        return cuda_version
    return "cpu"


def _run_ssh_command(
    address: str,
    ssh_user: str,
    ssh_port: int,
    ssh_private_key: str,
    ssh_proxy_command: str,
    docker_user: str,
):
    from runhouse.resources.hardware.sky_command_runner import SkySSHRunner

    runner = SkySSHRunner(
        (address, ssh_port),
        ssh_user=ssh_user,
        ssh_private_key=ssh_private_key,
        ssh_proxy_command=ssh_proxy_command,
        docker_user=docker_user,
    )
    ssh_command = runner._ssh_base_command(
        ssh_mode=SshMode.INTERACTIVE, port_forward=None
    )
    subprocess.run(ssh_command)


def _docker_ssh_proxy_command(
    address: str,
    ssh_user: str,
    ssh_private_key: str,
):
    return lambda ssh: " ".join(
        ssh
        + ssh_options_list(ssh_private_key, None)
        + ["-W", "%h:%p", f"{ssh_user}@{address}"]
    )


# Adapted from SkyPilot Command Runner
def _ssh_base_command(
    address: str,
    ssh_user: str,
    ssh_private_key: str,
    ssh_control_name: Optional[str] = "__default__",
    ssh_proxy_command: Optional[str] = None,
    ssh_port: int = 22,
    docker_ssh_proxy_command: Optional[str] = None,
    disable_control_master: Optional[bool] = False,
    ssh_mode: SshMode = SshMode.INTERACTIVE,
    port_forward: Optional[List[int]] = None,
    connect_timeout: Optional[int] = 30,
):
    ssh = ["ssh"]
    if ssh_mode == SshMode.NON_INTERACTIVE:
        # Disable pseudo-terminal allocation. Otherwise, the output of
        # ssh will be corrupted by the user's input.
        ssh += ["-T"]
    else:
        # Force pseudo-terminal allocation for interactive/login mode.
        ssh += ["-tt"]

    if port_forward is not None:
        # RH MODIFIED: Accept port int (to forward same port) or pair of ports
        for fwd in port_forward:
            if isinstance(fwd, int):
                local, remote = fwd, fwd
            else:
                local, remote = fwd
            logger.debug(f"Forwarding port {local} to port {remote} on localhost.")
            ssh += ["-L", f"{local}:localhost:{remote}"]

    return (
        ssh
        + ssh_options_list(
            ssh_private_key,
            ssh_control_name,
            ssh_proxy_command=ssh_proxy_command,
            docker_ssh_proxy_command=docker_ssh_proxy_command,
            # TODO change to None like before?
            port=ssh_port,
            connect_timeout=connect_timeout,
            disable_control_master=disable_control_master,
        )
        + [f"{ssh_user}@{address}"]
    )


def _generate_ssh_control_hash(ssh_control_name):
    return hashlib.md5(ssh_control_name.encode()).hexdigest()[:_HASH_MAX_LENGTH]


def up_cluster_helper(cluster, capture_output: Union[bool, str] = True):
    from runhouse.utils import SuppressStd

    if capture_output:
        try:
            with SuppressStd() as outfile:
                cluster.up()
        except Exception as e:
            if isinstance(capture_output, str):
                logger.error(
                    f"Error starting cluster {cluster.name}, logs written to {capture_output}"
                )
            raise e
        finally:
            if isinstance(capture_output, str):
                with open(capture_output, "w") as f:
                    f.write(outfile.output)
    else:
        cluster.up()



import ast
import datetime
import json
import os
import uuid
from enum import Enum

from requests import Response


def timing(func):
    def wrapper(*args, **kwargs):
        import time

        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(f"Finished {func.__name__.title()} in {int((end - start))} seconds")
        return result

    return wrapper


def remove_null_values_from_dict(source_dic: dict) -> dict:
    return {k: v for k, v in source_dic.items() if v is not None}


def load_resp_content(resp: Response) -> dict:
    return json.loads(resp.content)


def read_resp_data(resp: Response):
    return load_resp_content(resp).get("data", {})


def to_bool(value):
    try:
        return ast.literal_eval(value)
    except:
        return value


def is_jsonable(myjson):
    try:
        json.dumps(myjson)
    except (TypeError, OverflowError):
        return False
    return True


def generate_uuid():
    return uuid.uuid4().hex


def utc_now():
    """Current time as datetime object."""
    return datetime.datetime.now(datetime.timezone.utc)


def log_timestamp():
    """Return as timestamp in nanoseconds."""
    return int(utc_now().timestamp() * 1e9)


def log_datetime():
    """Current time as readable datetime string.
    Example: '2023-04-23'"""
    return utc_now().strftime("%Y-%m-%d")


def resolve_absolute_path(path: str):
    return os.path.abspath(os.path.expanduser(path))


def relative_file_path(file_path: str):
    """Convert to a relative path if it is not already one."""
    if file_path.startswith("~"):
        return file_path

    # Convert to a relative path
    relative_path = os.path.relpath(file_path, os.path.expanduser("~"))
    relative_path = relative_path.replace("\\", "/")

    if not relative_path.startswith("~"):
        relative_path = f"~/{relative_path}"

    return relative_path


class ResourceAccess(str, Enum):
    WRITE = "write"
    READ = "read"
    DENIED = "denied"


class ResourceVisibility(str, Enum):
    PRIVATE = "private"
    UNLISTED = "unlisted"
    PUBLIC = "public"



import argparse
import asyncio
import inspect
import json
import traceback
import uuid
from functools import wraps
from pathlib import Path
from typing import Optional

import ray
import requests
import yaml
from fastapi import Body, FastAPI, HTTPException, Request
from fastapi.encoders import jsonable_encoder
from fastapi.openapi.docs import get_redoc_html, get_swagger_ui_html
from fastapi.responses import StreamingResponse

from runhouse.constants import (
    DEFAULT_HTTP_PORT,
    DEFAULT_HTTPS_PORT,
    DEFAULT_SERVER_HOST,
    DEFAULT_SERVER_PORT,
    EMPTY_DEFAULT_ENV_NAME,
    LOGGING_WAIT_TIME,
    RH_LOGFILE_PATH,
)
from runhouse.globals import configs, obj_store, rns_client
from runhouse.logger import get_logger
from runhouse.rns.utils.api import resolve_absolute_path, ResourceAccess
from runhouse.servers.caddy.config import CaddyConfig
from runhouse.servers.http.auth import averify_cluster_access
from runhouse.servers.http.certs import TLSCertConfig
from runhouse.servers.http.http_utils import (
    CallParams,
    DeleteObjectParams,
    deserialize_data,
    folder_exists,
    folder_get,
    folder_ls,
    folder_mkdir,
    folder_mv,
    folder_put,
    folder_rm,
    FolderGetParams,
    FolderLsParams,
    FolderMvParams,
    FolderParams,
    FolderPutParams,
    FolderRmParams,
    get_token_from_request,
    handle_exception_response,
    OutputType,
    PutObjectParams,
    PutResourceParams,
    RenameObjectParams,
    resolve_folder_path,
    Response,
    serialize_data,
    ServerSettings,
)
from runhouse.servers.obj_store import (
    ClusterServletSetupOption,
    ObjStoreError,
    RaySetupOption,
)
from runhouse.utils import generate_default_name, sync_function

app = FastAPI(docs_url=None, redoc_url=None)

logger = get_logger(__name__)


def validate_cluster_access(func):
    """If using Den auth, validate the user's cluster subtoken and access to the cluster before continuing."""

    @wraps(func)
    async def wrapper(*args, **kwargs):
        request: Request = kwargs.get("request")
        den_auth_enabled: bool = await HTTPServer.get_den_auth()
        is_coro = inspect.iscoroutinefunction(func)

        func_call: bool = func.__name__ in ["post_call", "get_call"]

        # restrict access for folder specific APIs
        access_level_required = (
            ResourceAccess.WRITE if func.__name__.startswith("folder") else None
        )
        token = get_token_from_request(request)

        request_id = request.headers.get("X-Request-ID", str(uuid.uuid4()))
        ctx_token = obj_store.set_ctx(request_id=request_id, token=token)

        try:
            if den_auth_enabled and not func_call:
                if token is None:
                    raise HTTPException(
                        status_code=401,
                        detail="No Runhouse token provided. Try running `$ runhouse login` or visiting "
                        "https://run.house/login to retrieve a token. If calling via HTTP, please "
                        "provide a valid token in the Authorization header.",
                    )
                cluster_uri = (await obj_store.aget_cluster_config()).get("name")
                cluster_access = await averify_cluster_access(
                    cluster_uri, token, access_level_required
                )
                if not cluster_access:
                    # Must have cluster access for all the non func calls
                    # Note: for func calls we handle the auth in the object store
                    raise HTTPException(
                        status_code=403,
                        detail="Cluster access is required for this operation.",
                    )

            if is_coro:
                res = await func(*args, **kwargs)
            else:
                res = func(*args, **kwargs)
        except Exception as e:
            raise e
        finally:
            if ctx_token:
                obj_store.unset_ctx(ctx_token)
        return res

    return wrapper


class HTTPServer:
    SKY_YAML = str(Path("~/.sky/sky_ray.yml").expanduser())
    memory_exporter = None

    @classmethod
    async def ainitialize(
        cls,
        default_env_name=None,
        conda_env=None,
        from_test: bool = False,
        *args,
        **kwargs,
    ):
        runtime_env = {"conda": conda_env} if conda_env else None

        default_env_name = default_env_name or EMPTY_DEFAULT_ENV_NAME

        # Ray and ClusterServlet should already be
        # initialized by the start script (see below)
        # But if the HTTPServer was started standalone in a test,
        # We still want to make sure the cluster servlet is initialized
        if from_test:
            await obj_store.ainitialize(
                default_env_name,
                setup_ray=RaySetupOption.TEST_PROCESS,
                runtime_env=runtime_env,
            )

        # TODO disabling due to latency, figure out what to do with this
        # try:
        #     # Collect metadata for the cluster immediately on init
        #     self._collect_cluster_stats()
        # except Exception as e:
        #     logger.error(f"Failed to collect cluster stats: {str(e)}")

        # We initialize a default env servlet where some things may run.
        _ = obj_store.get_env_servlet(
            env_name=default_env_name,
            create=True,
            runtime_env=runtime_env,
        )

        if default_env_name == EMPTY_DEFAULT_ENV_NAME:
            from runhouse.resources.envs import Env

            default_env = Env(name=default_env_name)
            data = (default_env.config(condensed=False), {}, False)
            obj_store.put_resource(
                serialized_data=data, serialization=None, env_name=default_env_name
            )

    @classmethod
    def initialize(
        cls,
        default_env_name=None,
        conda_env=None,
        from_test: bool = False,
        *args,
        **kwargs,
    ):
        return sync_function(cls.ainitialize)(
            default_env_name,
            conda_env,
            from_test,
            *args,
            **kwargs,
        )

    @classmethod
    async def get_den_auth(cls):
        return obj_store.is_den_auth_enabled()

    @classmethod
    async def aenable_den_auth(cls, flush: Optional[bool] = True):
        await obj_store.aenable_den_auth()

        if flush:
            await obj_store.aclear_auth_cache()

    def enable_den_auth(flush: Optional[bool] = True):
        return sync_function(HTTPServer.aenable_den_auth)(flush)

    @classmethod
    async def adisable_den_auth(cls):
        await obj_store.adisable_den_auth()

    @classmethod
    def disable_den_auth(cls):
        return sync_function(HTTPServer.adisable_den_auth)()

    @staticmethod
    @app.get("/cert")
    def get_cert():
        """Download the certificate file for this server necessary for enabling HTTPS.
        User must have access to the cluster in order to download the certificate."""

        # Default for this endpoint is "pickle" serialization
        serialization = "pickle"

        try:
            certs_config = TLSCertConfig()
            cert_path = certs_config.cert_path

            if not Path(cert_path).exists():
                raise FileNotFoundError(
                    f"No certificate found on cluster in path: {cert_path}"
                )

            with open(cert_path, "rb") as cert_file:
                cert = cert_file.read()

            return Response(
                data=serialize_data(cert, serialization),
                output_type=OutputType.RESULT_SERIALIZED,
                serialization=serialization,
            )

        except Exception as e:
            logger.exception(e)
            exception_data = {
                "error": serialize_data(e, serialization),
                "traceback": traceback.format_exc(),
            }
            return Response(
                output_type=OutputType.EXCEPTION,
                data=exception_data,
                serialization=serialization,
            )

    @staticmethod
    @app.get("/check")
    def check_server():

        # Default for this endpoint is "pickle" serialization
        serialization = "pickle"

        try:
            if not ray.is_initialized():
                raise Exception("Ray is not initialized, restart the server.")
            logger.info("Server is up.")

            import runhouse

            return {"rh_version": runhouse.__version__}
        except Exception as e:
            logger.exception(e)
            exception_data = {
                "error": serialize_data(e, serialization),
                "traceback": traceback.format_exc(),
            }
            return Response(
                output_type=OutputType.EXCEPTION,
                data=exception_data,
                serialization=serialization,
            )

    @staticmethod
    @app.post("/settings")
    @validate_cluster_access
    async def update_settings(request: Request, message: ServerSettings) -> Response:
        if message.cluster_name:
            await obj_store.aset_cluster_config_value("name", message.cluster_name)

        if message.den_auth:
            await HTTPServer.aenable_den_auth(flush=message.flush_auth_cache)
        elif message.den_auth is not None and not message.den_auth:
            await HTTPServer.adisable_den_auth()

        if message.autostop_mins:
            obj_store.set_cluster_config_value("autostop_mins", message.autostop_mins)

        if message.status_check_interval:
            obj_store.set_cluster_config_value(
                "status_check_interval", message.status_check_interval
            )

        return Response(output_type=OutputType.SUCCESS)

    @staticmethod
    async def _call(
        key: str,
        method_name: Optional[str] = None,
        params: CallParams = Body(default=None),
    ):
        # Allow an empty body
        params = params or CallParams()

        try:
            if not params.run_name:
                raise ValueError("run_name is required for all calls.")
            # Call async so we can loop to collect logs until the result is ready

            fut = asyncio.create_task(
                obj_store.acall(
                    key=key,
                    method_name=method_name,
                    data=params.data,
                    stream_logs=params.stream_logs,
                    serialization=params.serialization,
                    run_name=params.run_name,
                    remote=params.remote,
                )
            )
            # If stream_logs is False, we'll wait for the result and return it
            if not params.stream_logs:
                return await fut

            return StreamingResponse(
                HTTPServer._get_results_and_logs_generator(
                    fut=fut,
                    run_name=params.run_name,
                    serialization=params.serialization,
                ),
                media_type="application/json",
            )
        except Exception as e:
            return handle_exception_response(
                e,
                traceback.format_exc(),
                serialization=params.serialization,
                from_http_server=True,
            )

    # Open API docs routes
    @staticmethod
    @app.get("/{key}/openapi.json")
    @validate_cluster_access
    async def get_openapi_spec(request: Request, key: str):
        try:
            module_openapi_spec = await obj_store.acall(
                key, method_name="openapi_spec", serialization=None
            )
        except (AttributeError, ObjStoreError):
            # The object put on the server is not an `rh.Module`, so it doesn't have an openapi_spec method
            # OR
            # The object is not found in the object store at all
            module_openapi_spec = None

        if not module_openapi_spec:
            raise HTTPException(status_code=404, detail=f"Module {key} not found.")

        return module_openapi_spec

    @staticmethod
    @app.get("/{key}/redoc")
    @validate_cluster_access
    async def get_redoc(request: Request, key: str):
        return get_redoc_html(
            openapi_url=f"/{key}/openapi.json", title="Developer Documentation"
        )

    @staticmethod
    @app.get("/{key}/docs")
    @validate_cluster_access
    async def get_swagger_ui_html(request: Request, key: str):
        return get_swagger_ui_html(
            openapi_url=f"/{key}/openapi.json",
            title=f"{key} - Swagger UI",
            oauth2_redirect_url=app.swagger_ui_oauth2_redirect_url,
            init_oauth=app.swagger_ui_init_oauth,
        )

    # TODO match "/{key}/{method_name}/{path:more_path}" for asgi / proxy requests
    @staticmethod
    @app.post("/{key}/{method_name}")
    @validate_cluster_access
    async def post_call(
        request: Request,
        key: str,
        method_name: str = None,
        params: CallParams = Body(default=None),
    ):
        return await HTTPServer._call(key, method_name, params)

    @staticmethod
    @app.get("/{key}/{method_name}")
    @validate_cluster_access
    async def get_call(
        request: Request,
        key: str,
        method_name: Optional[str] = None,
        serialization: Optional[str] = None,
        run_name: Optional[str] = None,
        stream_logs: Optional[bool] = False,
        save: Optional[bool] = False,
        remote: Optional[bool] = False,
    ):
        # Default argument to json doesn't allow a user to pass in a serialization string if they want
        # But, if they didn't pass anything, we want it to be `json` by default.
        serialization = serialization or "json"
        try:
            if run_name is None and stream_logs:
                raise ValueError(
                    "run_name is required for all calls when stream_logs is True."
                )

            if run_name is None:
                run_name = generate_default_name(
                    prefix=key if method_name == "__call__" else f"{key}_{method_name}",
                    precision="ms",  # Higher precision because we see collisions within the same second
                    sep="@",
                )

            # The types need to be explicitly specified as parameters first so that
            # we can cast Query params to the right type.
            params = CallParams(
                serialization=serialization,
                run_name=run_name,
                stream_logs=stream_logs,
                save=save,
                remote=remote,
            )

            query_params_remaining = dict(request.query_params)
            call_params_dict = params.model_dump()
            for k, v in dict(request.query_params).items():
                # If one of the query_params matches an arg in CallParams, set it
                # And also remove it from the query_params dict, so the rest
                # of the args will be passed as kwargs
                if k in call_params_dict:
                    del query_params_remaining[k]

            data = {
                "args": [],
                "kwargs": query_params_remaining,
            }
            params.data = serialize_data(data, serialization)

            return await HTTPServer._call(key, method_name, params)
        except Exception as e:
            logger.exception(e)
            exception_data = {
                "error": serialize_data(e, serialization),
                "traceback": traceback.format_exc(),
            }
            return Response(
                output_type=OutputType.EXCEPTION,
                data=exception_data,
                serialization=serialization,
            )

    @staticmethod
    def _get_logfiles(log_key, log_type=None):
        if not log_key:
            return None
        key_logs_path = Path(RH_LOGFILE_PATH) / log_key
        if key_logs_path.exists():
            # Logs are like: `.rh/logs/key/key.[out|err]`
            glob_pattern = (
                "*.out"
                if log_type == "stdout"
                else "*.err"
                if log_type == "stderr"
                else "*.[oe][ur][tr]"
            )
            return [str(f.absolute()) for f in key_logs_path.glob(glob_pattern)]
        else:
            return None

    @staticmethod
    def open_new_logfiles(key, open_files):
        logfiles = HTTPServer._get_logfiles(key)
        if logfiles:
            for f in logfiles:
                if f not in [o.name for o in open_files]:
                    logger.info(f"Streaming logs from {f}")
                    open_files.append(open(f, "r"))
        return open_files

    @staticmethod
    async def _get_results_and_logs_generator(fut, run_name, serialization=None):
        logger.debug(f"Streaming logs for key {run_name}")
        open_logfiles = []
        waiting_for_results = True

        try:
            while waiting_for_results:
                if fut.done():
                    waiting_for_results = False
                    ret_val = fut.result()
                    yield json.dumps(jsonable_encoder(ret_val)) + "\n"
                else:
                    await asyncio.sleep(LOGGING_WAIT_TIME)
                # Grab all the lines written to all the log files since the last time we checked, including
                # any new log files that have been created
                open_logfiles = HTTPServer.open_new_logfiles(run_name, open_logfiles)
                ret_lines = []
                for i, f in enumerate(open_logfiles):
                    file_lines = f.readlines()
                    if file_lines:
                        # TODO [DG] handle .out vs .err, and multiple workers
                        # if len(logfiles) > 1:
                        #     ret_lines.append(f"Process {i}:")
                        ret_lines += file_lines
                if ret_lines:
                    logger.debug(f"Yielding logs for key {run_name}")
                    yield json.dumps(
                        jsonable_encoder(
                            Response(
                                data=ret_lines,
                                output_type=OutputType.STDOUT,
                            )
                        )
                    ) + "\n"

        except Exception as e:
            logger.exception(e)
            # NOTE: We do not convert the exception to an HTTPException here, because once we're inside this
            # generator starlette has already returned the StreamingResponse and there is no way to halt the stream
            # to return a 403 instead. Users shouldn't notice much of a difference, because if working through the
            # client the exception will be serialized and returned to appear as a native python exception, and if
            # working through an HTTP call stream_logs is False by default, so a normal HTTPException will be raised
            # above before entering this generator.
            exception_data = {
                "error": serialize_data(e, serialization),
                "traceback": traceback.format_exc(),
            }
            yield json.dumps(
                jsonable_encoder(
                    Response(
                        output_type=OutputType.EXCEPTION,
                        data=exception_data,
                        serialization=serialization,
                    )
                )
            )
        finally:
            if not open_logfiles:
                logger.warning(f"No logfiles found for call {run_name}")
            for f in open_logfiles:
                f.close()

    @staticmethod
    @app.post("/resource")
    @validate_cluster_access
    async def put_resource(request: Request, params: PutResourceParams):
        try:
            env_name = params.env_name
            return await obj_store.aput_resource(
                serialized_data=params.serialized_data,
                serialization=params.serialization,
                env_name=env_name,
            )
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/object")
    @validate_cluster_access
    async def put_object(request: Request, params: PutObjectParams):
        try:
            await obj_store.aput(
                key=params.key,
                value=params.serialized_data,
                env=params.env_name,
                serialization=params.serialization,
                create_env_if_not_exists=True,
            )
            return Response(output_type=OutputType.SUCCESS)
        except Exception as e:
            return handle_exception_response(
                e,
                traceback.format_exc(),
                serialization=params.serialization,
                from_http_server=True,
            )

    @staticmethod
    @app.get("/object")
    @validate_cluster_access
    async def get_object(
        request: Request,
        key: str,
        serialization: Optional[str] = "json",
        remote: bool = False,
    ):
        try:
            return await obj_store.aget(
                key=key,
                serialization=serialization,
                remote=remote,
            )
        except Exception as e:
            return handle_exception_response(
                e,
                traceback.format_exc(),
                serialization=serialization,
                from_http_server=True,
            )

    @staticmethod
    @app.post("/rename")
    @validate_cluster_access
    async def rename_object(request: Request, params: RenameObjectParams):
        try:
            await obj_store.arename(
                old_key=params.key,
                new_key=params.new_key,
            )
            return Response(output_type=OutputType.SUCCESS)
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/folder/method/ls")
    @validate_cluster_access
    async def folder_ls_cmd(request: Request, ls_params: FolderLsParams):
        try:
            path = resolve_folder_path(ls_params.path)
            return folder_ls(path, full_paths=ls_params.full_paths, sort=ls_params.sort)

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/folder/method/mkdir")
    @validate_cluster_access
    async def folder_mkdir_cmd(request: Request, folder_params: FolderParams):
        try:
            path = resolve_folder_path(folder_params.path)
            return folder_mkdir(path)

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/folder/method/get")
    @validate_cluster_access
    async def folder_get_cmd(request: Request, get_params: FolderGetParams):
        try:
            path = resolve_folder_path(get_params.path)
            return folder_get(path, mode=get_params.mode, encoding=get_params.encoding)

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/folder/method/put")
    @validate_cluster_access
    async def folder_put_cmd(request: Request, put_params: FolderPutParams):
        try:
            path = resolve_folder_path(put_params.path)
            serialization = put_params.serialization
            serialized_contents = put_params.contents
            contents = deserialize_data(serialized_contents, serialization)

            return folder_put(
                path,
                contents=contents,
                overwrite=put_params.overwrite,
                mode=put_params.mode,
                serialization=serialization,
            )

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/folder/method/rm")
    @validate_cluster_access
    async def folder_rm_cmd(request: Request, rm_params: FolderRmParams):
        try:
            path = resolve_folder_path(rm_params.path)
            return folder_rm(
                path, contents=rm_params.contents, recursive=rm_params.recursive
            )

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/folder/method/mv")
    @validate_cluster_access
    async def folder_mv_cmd(request: Request, mv_params: FolderMvParams):
        try:
            path = resolve_folder_path(mv_params.path)
            return folder_mv(
                src_path=path,
                dest_path=mv_params.dest_path,
                overwrite=mv_params.overwrite,
            )

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/folder/method/exists")
    @validate_cluster_access
    async def folder_exists_cmd(request: Request, folder_params: FolderParams):
        try:
            path = resolve_folder_path(folder_params.path)
            return folder_exists(path=path)

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/delete_object")
    @validate_cluster_access
    async def delete_obj(request: Request, params: DeleteObjectParams):
        try:
            if len(params.keys) == 0:
                cleared = await obj_store.akeys()
                await obj_store.aclear()
            else:
                cleared = []
                for key in params.keys:
                    await obj_store.adelete(key)
                    cleared.append(key)

            # Expicitly tell the client not to attempt to deserialize the output
            return Response(
                data=cleared,
                output_type=OutputType.RESULT_SERIALIZED,
                serialization=None,
            )
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.get("/keys")
    @validate_cluster_access
    async def get_keys(request: Request, env_name: Optional[str] = None):
        try:
            if not env_name:
                output = await obj_store.akeys()
            else:
                output = await obj_store.akeys_for_env_servlet_name(env_name)

            # Expicitly tell the client not to attempt to deserialize the output
            return Response(
                data=output,
                output_type=OutputType.RESULT_SERIALIZED,
                serialization=None,
            )
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.get("/status")
    @validate_cluster_access
    def get_status(request: Request, send_to_den: bool = False):

        return obj_store.status(send_to_den=send_to_den)

    @staticmethod
    def _collect_cluster_stats():
        """Collect cluster metadata and send to Grafana Loki"""
        if not configs.data_collection_enabled():
            return

        cluster_data = HTTPServer._cluster_status_report()
        sky_data = HTTPServer._cluster_sky_report()

        HTTPServer._log_cluster_data(
            {**cluster_data, **sky_data},
            labels={"username": configs.username, "environment": "prod"},
        )

    @staticmethod
    def _cluster_status_report():
        import ray._private.usage.usage_lib as ray_usage_lib
        from ray._raylet import GcsClient

        gcs_client = GcsClient(address="127.0.0.1:6379", nums_reconnect_retry=20)

        # fields : ['ray_version', 'python_version']
        cluster_metadata = ray_usage_lib.get_cluster_metadata(gcs_client)

        # fields: ['total_num_cpus', 'total_num_gpus', 'total_memory_gb', 'total_object_store_memory_gb']
        cluster_status_report = ray_usage_lib.get_cluster_status_to_report(
            gcs_client
        ).__dict__

        return {**cluster_metadata, **cluster_status_report}

    @staticmethod
    def _cluster_sky_report():
        try:
            with open(HTTPServer.SKY_YAML, "r") as stream:
                sky_ray_data = yaml.safe_load(stream)
        except FileNotFoundError:
            # For non on-demand clusters we won't have sky data
            return {}

        provider = sky_ray_data["provider"]
        node_config = sky_ray_data["available_node_types"].get("ray.head.default", {})

        return {
            "cluster_name": sky_ray_data.get("cluster_name"),
            "region": provider.get("region"),
            "provider": provider.get("module"),
            "instance_type": node_config.get("node_config", {}).get("InstanceType"),
        }

    @staticmethod
    def _log_cluster_data(data: dict, labels: dict):
        from runhouse.rns.utils.api import log_timestamp

        payload = {
            "streams": [
                {"stream": labels, "values": [[str(log_timestamp()), json.dumps(data)]]}
            ]
        }

        payload = json.dumps(payload)
        resp = requests.post(
            f"{rns_client.api_server_url}/admin/logs", data=json.dumps(payload)
        )

        if resp.status_code == 405:
            # api server not configured to receive grafana logs
            return

        if resp.status_code != 200:
            logger.error(
                f"({resp.status_code}) Failed to send logs to Grafana Loki: {resp.text}"
            )


async def main():
    import uvicorn

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--host",
        type=str,
        default=None,
        help=f"Host to run server on. By default will run on {DEFAULT_SERVER_HOST}",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=None,
        help="Port to run daemon on on. If provided and Caddy is not enabled, "
        f"will attempt to run the daemon on this port, defaults to {DEFAULT_SERVER_PORT}",
    )
    parser.add_argument(
        "--conda-env", type=str, default=None, help="Conda env to run server in"
    )
    parser.add_argument(
        "--use-https",
        action="store_true",  # if providing --use-https will be set to True
        default=argparse.SUPPRESS,  # If user didn't specify, attribute will not be present (not False)
        help="Start an HTTPS server with new TLS certs",
    )
    parser.add_argument(
        "--use-den-auth",
        action="store_true",  # if providing --use-den-auth will be set to True
        default=argparse.SUPPRESS,  # If user didn't specify, attribute will not be present (not False)
        help="Whether to authenticate requests with a Runhouse token",
    )
    parser.add_argument(
        "--domain",
        type=str,
        default=None,
        help="Server domain name",
    )
    parser.add_argument(
        "--ssl-keyfile",
        type=str,
        default=None,
        help="Path to SSL key file on the cluster to use for HTTPS",
    )
    parser.add_argument(
        "--ssl-certfile",
        type=str,
        default=None,
        help="Path to SSL cert file on the cluster to use for HTTPS",
    )
    parser.add_argument(
        "--restart-proxy",
        action="store_true",  # if providing --restart-proxy will be set to True
        help="Reconfigure Caddy",
    )
    parser.add_argument(
        "--use-caddy",
        action="store_true",  # if providing --use-caddy will be set to True
        default=argparse.SUPPRESS,  # If user didn't specify, attribute will not be present (not False)
        help="Configure Caddy as a reverse proxy",
    )
    parser.add_argument(
        "--certs-address",
        type=str,
        default=None,
        help="Address to use for generating self-signed certs and enabling HTTPS. (e.g. public IP address)",
    )
    parser.add_argument(
        "--default-env-name",
        type=str,
        default=None,
        help="Name of env where the HTTP server is started.",
    )
    parser.add_argument(
        "--api-server-url",
        type=str,
        default=rns_client.api_server_url,
        help="URL of Runhouse Den",
    )
    parser.add_argument(
        "--from-python",
        action="store_true",
        default=argparse.SUPPRESS,
        help="Whether HTTP server is called from Python rather than CLI.",
    )

    parse_args = parser.parse_args()

    conda_name = parse_args.conda_env
    restart_proxy = parse_args.restart_proxy
    api_server_url = parse_args.api_server_url
    default_env_name = parse_args.default_env_name

    if not hasattr(parse_args, "from_python") and not default_env_name:
        # detect default env if called from runhouse cli (start/restart) and cluster_config exists
        from runhouse.resources.hardware.utils import load_cluster_config_from_file

        cluster_config = load_cluster_config_from_file()
        if cluster_config.get("default_env"):
            default_env_name = cluster_config.get("default_env")["name"]

    # The object store and the cluster servlet within it need to be
    # initialized in order to call `obj_store.get_cluster_config()`, which
    # uses the object store to load the cluster config from Ray.
    # When setting up the server, we always want to create a new ClusterServlet.
    # We only want to forcibly start a Ray cluster if asked.
    # We connect this to the "base" env, which we'll initialize later,
    # so writes to the obj_store within the server get proxied to the "base" env.
    await obj_store.ainitialize(
        default_env_name,
        setup_cluster_servlet=ClusterServletSetupOption.FORCE_CREATE,
    )

    cluster_config = await obj_store.aget_cluster_config()
    if not cluster_config:
        logger.warning(
            "Cluster config is not set. Using default values where possible."
        )
    else:
        logger.info("Loaded cluster config from Ray.")

    ########################################
    # Handling args that could be specified in the
    # cluster_config.json or via CLI args
    ########################################

    # Den URL
    cluster_config["api_server_url"] = api_server_url

    # Server port
    if parse_args.port is not None and parse_args.port != cluster_config.get(
        "server_port"
    ):
        logger.warning(
            f"CLI provided server port: {parse_args.port} is different from the server port specified in "
            f"cluster_config.json: {cluster_config.get('server_port')}. Prioritizing CLI provided port."
        )

    port_arg = (
        parse_args.port or cluster_config.get("server_port") or DEFAULT_SERVER_PORT
    )

    if port_arg is not None:
        cluster_config["server_port"] = port_arg

    # Den auth enabled
    if hasattr(
        parse_args, "use_den_auth"
    ) and parse_args.use_den_auth != cluster_config.get("den_auth", False):
        logger.warning(
            f"CLI provided den_auth: {parse_args.use_den_auth} is different from the den_auth specified in "
            f"cluster_config.json: {cluster_config.get('den_auth')}. Prioritizing CLI provided den_auth."
        )

    den_auth = getattr(parse_args, "use_den_auth", False) or cluster_config.get(
        "den_auth", False
    )
    cluster_config["den_auth"] = den_auth

    if den_auth:
        await obj_store.aenable_den_auth()

    domain = parse_args.domain or cluster_config.get("domain", None)
    cluster_config["domain"] = domain

    # Keyfile
    if (
        parse_args.ssl_keyfile is not None
        and parse_args.ssl_keyfile != cluster_config.get("ssl_keyfile")
    ):
        logger.warning(
            f"CLI provided ssl_keyfile: {parse_args.ssl_keyfile} is different from the ssl_keyfile specified in "
            f"cluster_config.json: {cluster_config.get('ssl_keyfile')}. Prioritizing CLI provided ssl_keyfile."
        )

    ssl_keyfile_path_arg = parse_args.ssl_keyfile or cluster_config.get("ssl_keyfile")
    if ssl_keyfile_path_arg is not None:
        cluster_config["ssl_keyfile"] = ssl_keyfile_path_arg
    parsed_ssl_keyfile = (
        resolve_absolute_path(ssl_keyfile_path_arg) if ssl_keyfile_path_arg else None
    )

    # Certfile
    if (
        parse_args.ssl_certfile is not None
        and parse_args.ssl_certfile != cluster_config.get("ssl_certfile")
    ):
        logger.warning(
            f"CLI provided ssl_certfile: {parse_args.ssl_certfile} is different from the ssl_certfile specified in "
            f"cluster_config.json: {cluster_config.get('ssl_certfile')}. Prioritizing CLI provided ssl_certfile."
        )

    ssl_certfile_path_arg = parse_args.ssl_certfile or cluster_config.get(
        "ssl_certfile"
    )
    if ssl_certfile_path_arg is not None:
        cluster_config["ssl_certfile"] = ssl_certfile_path_arg
    parsed_ssl_certfile = (
        resolve_absolute_path(ssl_certfile_path_arg) if ssl_certfile_path_arg else None
    )

    # Host
    if parse_args.host is not None and parse_args.host != cluster_config.get(
        "server_host"
    ):
        logger.warning(
            f"CLI provided server_host: {parse_args.host} is different from the server_host specified in "
            f"cluster_config.json: {cluster_config.get('server_host')}. Prioritizing CLI provided server_host."
        )

    host = parse_args.host or cluster_config.get("server_host") or DEFAULT_SERVER_HOST
    cluster_config["server_host"] = host

    # Address in the case we're a TLS server
    if (
        parse_args.certs_address is not None
        and parse_args.certs_address != cluster_config.get("ips", [None])[0]
    ):
        logger.warning(
            f"CLI provided certs_address: {parse_args.certs_address} is different from the certs_address specified in "
            f"cluster_config.json: {cluster_config.get('ips', [None])[0]}. Prioritizing CLI provided certs_address."
        )

    certs_addresses = parse_args.certs_address or cluster_config.get("ips", None)

    # We don't want to unset multiple addresses if they were set in the cluster config
    if certs_addresses is not None:
        cluster_config["ips"] = certs_addresses
    else:
        cluster_config["ips"] = ["0.0.0.0"]

    certs_address = certs_addresses[0] if certs_addresses else None

    config_conn = cluster_config.get("server_connection_type")

    # Use caddy as reverse proxy
    if hasattr(parse_args, "use_caddy") and parse_args.use_caddy != (
        config_conn in ["tls", "none"]
    ):
        logger.warning(
            f"CLI provided use_caddy: {parse_args.use_caddy} is different from the server_connection_type specified in "
            f"cluster_config.json: {config_conn}. Prioritizing CLI provided use_caddy."
        )

    # Use HTTPS
    if hasattr(parse_args, "use_https") and parse_args.use_https != (
        config_conn == "tls"
    ):
        logger.warning(
            f"CLI provided use_https: {parse_args.use_https} is different from the server_connection_type specified in "
            f"cluster_config.json: {config_conn}. Prioritizing CLI provided use_https."
        )

    use_caddy = getattr(parse_args, "use_caddy", False) or (
        config_conn in ["tls", "none"]
    )
    use_https = getattr(parse_args, "use_https", False) or (config_conn == "tls")
    cluster_config["server_connection_type"] = (
        "tls" if use_https else "none" if use_caddy else config_conn
    )

    # If there was no `cluster_config.json`, then server was created
    # simply with `runhouse start`.
    # A real `cluster_config` that was loaded
    # from json would have this set for sure
    if not cluster_config.get("resource_subtype"):
        # This is needed for the Cluster object to be created
        # in rh.here.
        cluster_config["resource_subtype"] = "Cluster"

        # server_connection_type is not set up if this is done through
        # a local `runhouse start`
        cluster_config["server_connection_type"] = "tls" if use_https else "none"

    await obj_store.aset_cluster_config(cluster_config)

    await HTTPServer.ainitialize(
        default_env_name=default_env_name,
        conda_env=conda_name,
    )

    if den_auth:
        # Update den auth if enabled - keep as a class attribute to be referenced by the validator decorator
        await HTTPServer.aenable_den_auth()

    if use_https and not domain:
        # If using https (whether or not Caddy is being used) and no domain is specified, need to provide both
        # key and cert files - they should both already exist on the cluster
        if (
            not parsed_ssl_keyfile
            or not Path(parsed_ssl_keyfile).exists()
            or not parsed_ssl_certfile
            or not Path(parsed_ssl_certfile).exists()
        ):
            # Custom certs should already be on the cluster if their file paths are provided
            raise FileNotFoundError(
                f"Could not find SSL private key and cert files on the cluster, which are required when specifying "
                f"a port ({port_arg}) or enabling HTTPS. Please specify the paths using the --ssl-certfile and "
                f"--ssl-keyfile flags."
            )

    # If the daemon port was not specified, it should be the default RH port
    daemon_port = port_arg
    if not daemon_port or daemon_port in [
        DEFAULT_HTTP_PORT,
        DEFAULT_HTTPS_PORT,
    ]:
        # Since one of HTTP_PORT or HTTPS_PORT was specified, Caddy is set up to forward requests
        # from the daemon to that port
        daemon_port = DEFAULT_SERVER_PORT

    # Note: running the FastAPI app on a higher, non-privileged port (8000) and using Caddy as a reverse
    # proxy to forward requests from port 80 (HTTP) or 443 (HTTPS) to the app's port.
    if use_caddy:
        logger.debug("Using Caddy as a reverse proxy")
        if certs_address is None and domain is None:
            raise ValueError(
                "Must provide the server address or domain to configure Caddy. No address or domain found in the "
                "server start command (--certs-address or --domain) or in the cluster config YAML saved on the cluster."
            )

        cc = CaddyConfig(
            address=certs_address,
            domain=domain,
            rh_server_port=daemon_port,
            ssl_key_path=parsed_ssl_keyfile,
            ssl_cert_path=parsed_ssl_certfile,
            use_https=use_https,
            force_reinstall=restart_proxy,
        )
        cc.configure()

        caddy_port = DEFAULT_HTTPS_PORT if use_https else DEFAULT_HTTP_PORT
        logger.info(f"Caddy is proxying requests from {caddy_port} to {daemon_port}.")

    logger.info(
        f"Launching Runhouse API server with den_auth={den_auth} and "
        + f"on host={host} and use_https={use_https} and port_arg={daemon_port}"
    )

    cluster = None
    if not hasattr(parse_args, "from_python") and default_env_name:
        from runhouse.resources.hardware import Cluster

        cluster = Cluster.from_config(cluster_config)
        cluster._sync_default_env_to_cluster()

    # Only launch uvicorn with certs if HTTPS is enabled and not using Caddy
    uvicorn_cert = parsed_ssl_certfile if not use_caddy and use_https else None
    uvicorn_key = parsed_ssl_keyfile if not use_caddy and use_https else None

    config = uvicorn.Config(
        app,
        host=host,
        port=daemon_port,
        ssl_certfile=uvicorn_cert,
        ssl_keyfile=uvicorn_key,
        loop="uvloop",
    )
    server = uvicorn.Server(config)
    await server.serve()

    if cluster:
        cluster.put_resource(cluster.default_env)

        from runhouse.resources.envs.utils import _process_env_vars

        env_vars = _process_env_vars(cluster.default_env.env_vars)
        if env_vars:
            cluster.call(cluster.default_env.name, "_set_env_vars", env_vars)


if __name__ == "__main__":
    asyncio.run(main())



import base64
import contextlib
import copy
import importlib
import inspect
import json
import time
import warnings
import zipfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

try:
    import boto3
    import botocore.exceptions
except ImportError:
    pass

from runhouse.globals import rns_client
from runhouse.logger import get_logger
from runhouse.resources.envs import _get_env_from, Env
from runhouse.resources.functions.function import Function


CRED_PATH = f"{Path.home()}/.aws/credentials"
LOG_GROUP_PREFIX = "/aws/lambda/"

logger = get_logger(__name__)


class LambdaFunction(Function):
    RESOURCE_TYPE = "lambda_function"
    DEFAULT_ACCESS = "write"
    DEFAULT_ROLE_POLICIES = [
        "cloudwatch:*",
        "lambda:Invoke",
        "lambda:InvokeAsync",
        "lambda:InvokeFunction",
        "lambda:PublishVersion",
        "logs:*",
        "s3:DeleteObject",
        "s3:GetObject",
        "s3:ListBucket",
        "s3:PutObject",
        "kms:Decrypt",
    ]
    MAX_WAIT_TIME = 60  # seconds, max time that can pass before we raise an exception that AWS update takes too long.
    DEFAULT_REGION = "us-east-1"
    DEFAULT_RETENTION = 30  # one month, for lambdas log groups.
    DEFAULT_TIMEOUT = 900  # sec. meaning 15 min.
    DEFAULT_MEMORY_SIZE = 10240  # MB
    DEFAULT_TMP_SIZE = 10240  # MB, meaning 10G
    HOME_DIR = "/tmp/home"
    SUPPORTED_RUNTIMES = [
        "python3.7",
        "python3.8",
        "python3.9",
        "python3.10",
        "python3.11",
    ]
    DEFAULT_PY_VERSION = "python3.9"

    def __init__(
        self,
        paths_to_code: List[str],
        handler_function_name: str,
        runtime: str,
        fn_pointers: tuple,
        name: str,
        env: Env,
        timeout: int,  # seconds
        memory_size: int,  # MB
        tmp_size: int,  # MB
        retention_time: int = DEFAULT_RETENTION,  # days
        dryrun: bool = False,
        access: Optional[str] = None,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        Runhouse AWS Lambda object. It is comprised of the entry point, configuration,
        and dependencies necessary to run the service on AWS infra.

        .. note::
                To create an AWS lambda resource, please use the factory method :func:`aws_lambda_fn`.
        """

        super().__init__(
            name=name,
            dryrun=dryrun,
            system=self.RESOURCE_TYPE,
            env=env,
            fn_pointers=fn_pointers,
            access=access or self.DEFAULT_ACCESS,
            **kwargs,
        )

        self.local_path_to_code = paths_to_code
        self.handler_function_name = handler_function_name
        self.runtime = runtime
        self.timeout = timeout
        self.memory_size = memory_size
        self.tmp_size = tmp_size
        self.aws_lambda_config = (
            None  # Lambda config and role arn from aws will be saved here
        )
        self.retention_time = retention_time
        self.args_names = LambdaFunction.extract_args_from_file(
            paths_to_code, handler_function_name
        )

        # set-up in order to prevent read timeout during lambda invocations
        config = botocore.config.Config(read_timeout=900, connect_timeout=900)

        # will be used for reloading shared functions from other regions.
        function_arn = (
            kwargs.get("function_arn") if "function_arn" in kwargs.keys() else None
        )
        if function_arn:
            region = function_arn.split(":")[3]
            self.lambda_client = boto3.client(
                "lambda", region_name=region, config=config
            )
            self.logs_client = boto3.client("logs", region_name=region)
        else:

            if Path(CRED_PATH).is_file():
                self.lambda_client = boto3.client("lambda", config=config)
                self.logs_client = boto3.client("logs")

            else:
                self.lambda_client = boto3.client(
                    "lambda", region_name=self.DEFAULT_REGION
                )
                self.logs_client = boto3.client("logs", region_name=self.DEFAULT_REGION)
        self.iam_client = boto3.client("iam")

    # --------------------------------------
    # Constructor helper methods
    # --------------------------------------

    @classmethod
    def from_config(
        cls, config: dict, dryrun: bool = False, _resolve_children: bool = True
    ):
        if "resource_subtype" in config.keys():
            config.pop("resource_subtype", None)
        if "system" in config.keys():
            config.pop("system", None)
        keys = config.keys()
        if "fn_pointers" not in keys:
            config["fn_pointers"] = None
        if "timeout" not in keys:
            config["timeout"] = LambdaFunction.DEFAULT_TIMEOUT
        if "memory_size" not in keys:
            config["memory_size"] = LambdaFunction.DEFAULT_MEMORY_SIZE
        if "env" not in keys:
            config["env"] = Env(
                reqs=[],
                env_vars={"HOME": LambdaFunction.HOME_DIR},
            )
        else:
            config["env"] = Env.from_config(
                config["env"], _resolve_children=_resolve_children
            )

        return LambdaFunction(**config, dryrun=dryrun).deploy()

    @classmethod
    def from_name(
        cls,
        name,
        load_from_den: bool = True,
        dryrun: bool = False,
        _alt_options: Dict = None,
        _resolve_children: bool = True,
    ):
        config = rns_client.load_config(name=name, load_from_den=load_from_den)
        if not config:
            raise ValueError(f"Could not find a Lambda called {name}.")
        return cls.from_config(config, _resolve_children=_resolve_children)

    @classmethod
    def from_handler_file(
        cls,
        paths_to_code: List[str],
        handler_function_name: str,
        name: Optional[str] = None,
        env: Optional[Union[Dict, List[str], Env]] = None,
        runtime: Optional[str] = None,
        timeout: Optional[int] = None,
        memory_size: Optional[int] = None,
        tmp_size: Optional[int] = None,
        retention_time: Optional[int] = None,
        dryrun: bool = False,
    ):
        """
        Creates an AWS Lambda function from a Python file.

        Args:
            paths_to_code: (Optional[List[str]]): List of the FULL paths to the Python code file(s) that should be
                sent to AWS Lambda. First path in the list should be the path to the handler file which contains
                the main (handler) function. If ``fn`` is provided, this argument is ignored.
            handler_function_name: (Optional[str]): The name of the function in the handler file that will be executed
                by the Lambda. If ``fn`` is provided, this argument is ignored.
            runtime: (Optional[str]): The coding language of the function. Should be one of the following:
                python3.7, python3.8, python3.9, python3.10, python3.11. (Default: ``python3.9``)
            name (Optional[str]): Name of the Lambda Function to create or retrieve.
                This can be either from a local config or from the RNS.
            env (Optional[Dict or List[str] or Env]): Specifies the requirements that will be installed, and the
                environment vars that should be attached to the Lambda. Accepts three possible types:

                1. A dict which should contain the following keys:

                    reqs: a list of the python libraries, to be installed by the Lambda, or just a
                    ``requirements.txt``string.

                    env_vars: dictionary containing the env_vars that will be a part of the lambda configuration.

                2. A list of strings, containing all the required python packages.

                3. An instance of Runhouse Env class. By default, ``runhouse`` package will be installed, and env_vars
                   will include ``{HOME: /tmp/home}``.

            timeout (Optional[int]) The maximum amount of time (in seconds) during which the Lambda will run in AWS
                without timing-out. (Default: ``900``, Min: ``3``, Max: ``900``)
            memory_size (Optional[int]), The amount of memory (in MB) to be allocated to the Lambda.
                (Default: ``10240``, Min: ``128``, Max: ``10240``)
            tmp_size (Optional[int]), This size of the /tmp folder in the aws lambda file system.
                (Default: ``10240``, Min: ``512``, Max: ``10240``).
            retention_time (Optional[int]) The time (in days) the Lambda execution logs will be saved in AWS
                cloudwatch. After that, they will be deleted. (Default: ``30`` days)
            dryrun (bool): Whether to create the Function if it doesn't exist, or load the Function object as a dryrun.
                (Default: ``False``).

        Returns:
            LambdaFunction: The resulting AWS Lambda Function object.

        Example:
            >>> # handler_file.py
            >>> def summer(a, b):
            >>>    return a + b

            >>> # your 'main' python file, where you are using runhouse
            >>> summer_lambda = rh.LambdaFunction.from_handler_file(
            >>>                     paths_to_code=['/full/path/to/handler_file.py'],
            >>>                     handler_function_name = 'summer',
            >>>                     runtime = 'python3.9',
            >>>                     name="my_func").save()

            >>> # invoking the function
            >>> summer_res = summer_lambda(5, 8)  # returns 13.
        """

        if name is None:
            name = handler_function_name.replace(".", "_")

        env = LambdaFunction._validate_and_create_env(env)
        (
            paths_to_code,
            env,
            runtime,
            timeout,
            memory_size,
            tmp_size,
            retention_time,
        ) = LambdaFunction.arguments_validation(
            paths_to_code, env, runtime, timeout, memory_size, tmp_size, retention_time
        )

        new_function = LambdaFunction(
            fn_pointers=None,
            paths_to_code=paths_to_code,
            handler_function_name=handler_function_name,
            runtime=runtime,
            name=name,
            env=env,
            dryrun=dryrun,
            timeout=timeout,
            memory_size=memory_size,
            tmp_size=tmp_size,
            retention_time=retention_time,
        )

        if dryrun:
            return new_function

        return new_function.deploy()

    # --------------------------------------
    # Private helping methods
    # --------------------------------------

    # Arguments validation and arguments creation methods
    @classmethod
    def _validate_and_create_env(cls, env: Optional[Env] = None):
        """
        Validates the passed env argument, and creates a Runhouse env instance if needed.

        Args:
            env (Optional[Env]): The environment to validate and create.

        Returns:
            Env: a Runhouse env object
        """
        if env is not None and not isinstance(env, Env):
            original_env = copy.deepcopy(env)
            if isinstance(original_env, dict) and "env_vars" in original_env.keys():
                env = _get_env_from(env["reqs"]) or Env(working_dir="./")
            elif isinstance(original_env, str):
                env = _get_env_from(env)
            else:
                env = _get_env_from(env) or Env(working_dir="./")

        if env is None:
            env = Env(
                reqs=[],
                env_vars={"HOME": cls.HOME_DIR},
            )

        if isinstance(env, Env) and "HOME" not in env.env_vars.keys():
            env.env_vars["HOME"] = cls.HOME_DIR

        return env

    @classmethod
    def extract_args_from_file(cls, paths_to_code, handler_function_name):
        file_path = paths_to_code[0]
        name = file_path.split("/")[-1].split(".")[0]

        spec = importlib.util.spec_from_file_location(name, file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        # Extract the function and its arguments
        func = getattr(module, handler_function_name, None)
        args_names = inspect.getfullargspec(func).args
        warnings.warn(
            f"Arguments names were not provided. Extracted the following args names: {args_names}."
        )
        return args_names

    @classmethod
    def arguments_validation(
        cls, paths_to_code, env, runtime, timeout, memory_size, tmp_size, retention_time
    ):

        if isinstance(env, str) and "requirements.txt" in env:
            paths_to_code.append(Path(env).absolute())

        if runtime is None or runtime not in LambdaFunction.SUPPORTED_RUNTIMES:
            warnings.warn(
                f"{runtime} is not a supported by AWS Lambda. Setting runtime to python3.9."
            )
            runtime = LambdaFunction.DEFAULT_PY_VERSION

        if timeout is None:
            warnings.warn("Timeout set to 15 min.")
            timeout = LambdaFunction.DEFAULT_TIMEOUT
        else:
            if (env.reqs is not None or len(env.reqs) > 0) and timeout < 600:
                warnings.warn(
                    "Increasing the timeout to 600 sec, in order to enable the packages setup."
                )
                timeout = 600
            if timeout > 900:
                timeout = 900
                warnings.warn(
                    "Timeout can not be more then 900 sec, setting to 900 sec."
                )
            if timeout < 3:
                timeout = 3
                warnings.warn("Timeout can not be less then 3 sec, setting to 3 sec.")
        if memory_size is None:
            warnings.warn("Memory size set to 10 GB.")
            memory_size = LambdaFunction.DEFAULT_MEMORY_SIZE
        else:
            if (
                env.reqs is not None or len(env.reqs) > 0
            ) and memory_size < LambdaFunction.DEFAULT_MEMORY_SIZE:
                warnings.warn(
                    "Increasing the memory size to 10240 MB, in order to enable the packages setup."
                )
                memory_size = LambdaFunction.DEFAULT_MEMORY_SIZE
            if memory_size < 128:
                memory_size = 128
                warnings.warn(
                    "Memory size can not be less then 128 MB, setting to 128 MB."
                )
            if memory_size > 10240:
                memory_size = 10240
                warnings.warn(
                    "Memory size can not be more then 10240 MB, setting to 10240 MB."
                )
        if tmp_size is None or tmp_size < LambdaFunction.DEFAULT_TMP_SIZE:
            warnings.warn(
                "Setting /tmp size to 10GB, in order to enable the packages setup."
            )
            tmp_size = LambdaFunction.DEFAULT_TMP_SIZE
        elif tmp_size > 10240:
            tmp_size = 10240
            warnings.warn(
                "/tmp size can not be more then 10240 MB, setting to 10240 MB."
            )
        if retention_time is None:
            retention_time = LambdaFunction.DEFAULT_RETENTION

        return (
            paths_to_code,
            env,
            runtime,
            timeout,
            memory_size,
            tmp_size,
            retention_time,
        )

    def _lambda_exist(self, name):
        """Checks if a Lambda with the name given during init is already exists in AWS"""
        try:
            aws_lambda = self.lambda_client.get_function(FunctionName=name)
            return aws_lambda
        except self.lambda_client.exceptions.ResourceNotFoundException:
            return False

    @contextlib.contextmanager
    def _wait_until_lambda_update_is_finished(self, name):
        """Verifies that a running update of the function (in AWS) is finished (so the next one could be executed)"""

        time_passed = 0  # seconds
        response = self.lambda_client.get_function(FunctionName=name)
        state = response["Configuration"]["State"] == "Active"

        try:
            last_update_status = (
                response["Configuration"]["LastUpdateStatus"] != "InProgress"
            )
        except KeyError:
            last_update_status = False

        while True:
            if state and last_update_status:
                break
            else:
                if time_passed > self.MAX_WAIT_TIME:
                    raise TimeoutError(
                        f"Lambda function called {name} is being deployed in AWS for too long. "
                        + " Please check the resource in AWS console, delete relevant resource(s) "
                        + "if necessary, and re-run your Runhouse code."
                    )
                time.sleep(1)
                time_passed += 1
                response = self.lambda_client.get_function(FunctionName=name)
                state = response["Configuration"]["State"] == "Active"
                try:
                    last_update_status = (
                        response["Configuration"]["LastUpdateStatus"] != "InProgress"
                    )
                except KeyError:
                    last_update_status = False

        yield True

    def _rh_wrapper(self):
        """Creates a runhouse wrapper to the handler function"""
        handler_path = self.local_path_to_code[0]
        wrapper_path = str(Path(handler_path).parent / f"rh_handler_{self.name}.py")
        handler_name = Path(handler_path).stem
        Path(handler_path).touch()

        f = open(wrapper_path, "w")
        f.write("import subprocess\n" "import sys\n" "import os\n\n")
        f.write("def lambda_handler(event, context):\n")
        f.write(f"\tif not os.path.isdir('{self.HOME_DIR}'):\n")
        f.write(f"\t\tos.mkdir('{self.HOME_DIR}')\n")

        # adding code for installing python libraries
        if isinstance(self.env, str):
            f.write(
                "\tsubprocess.call(['pip', 'install', '-r', 'requirements.txt',"
                + " '--ignore-installed', '-t', '{self.HOME_DIR}/'])\n"
                f"\tif not os.path.isdir('{self.HOME_DIR}/runhouse'):\n"
                f"\t\tsubprocess.call(['pip', 'install', 'runhouse', '-t', '{self.HOME_DIR}/'])\n"
                f"\tsys.path.insert(1, '{self.HOME_DIR}/')\n\n"
            )
        else:
            reqs = self.env.reqs
            if "runhouse" not in reqs:
                reqs.append("runhouse")
            if "./" in reqs:
                reqs.remove("./")
            for req in reqs:
                req = req.split()
                dir_name = req[0]
                if len(req) > 1:
                    req = "".join("'%s', " % str(s) for s in req)[:-1]
                else:
                    req = f"'{req[0]}',"
                f.write(
                    f"\tif not os.path.isdir('{self.HOME_DIR}/{dir_name}'):\n"
                    f"\t\tsubprocess.call(['pip', 'install', {req} '-t', '{self.HOME_DIR}/'])\n"
                )
            if reqs is not None and len(reqs) > 0:
                f.write(f"\tsys.path.insert(1, '{self.HOME_DIR}/')\n\n")

        f.write(
            "\timport runhouse\n"
            f"\tfrom {handler_name} import {self.handler_function_name}\n"
            "\treturn {\n"
            "\t\t'status_code': 200,\n"
            f"\t\t'body': {self.handler_function_name}(**event)\n"
            "\t}"
        )
        f.close()
        return wrapper_path

    @classmethod
    def _supported_python_libs(cls):
        """Returns a list of the supported python libs by the AWS Lambda resource"""
        # TODO [SB]: think what is the better implementation: via website, or AWS lambda. For now it is hard-coded.
        supported_libs = [
            "urllib3",
            "six",
            "simplejson",
            "s3transfer",
            "python-dateutil",
            "jmespath",
            "botocore",
            "boto3",
            "awslambdaric",
            "setuptools",
            "pip",
        ]

        return supported_libs

    def _update_lambda_config(self, env_vars):
        """Updates existing Lambda in AWS (config) that was provided in the init."""
        logger.info(f"Updating a Lambda called {self.name}")
        lambda_config = self.lambda_client.update_function_configuration(
            FunctionName=self.name,
            Runtime=self.runtime,
            Timeout=self.timeout,
            MemorySize=self.memory_size,
            Environment={"Variables": env_vars},
        )

        # TODO [SB]: in the next phase, enable for other to update the Lambda code.
        # wait for the config update process to finish, and then update the code (Lambda logic).
        with self._wait_until_lambda_update_is_finished(self.name):
            path = str(Path(self.local_path_to_code[0]).absolute()).split("/")
            path = ["/" + path_e for path_e in path]
            path[0] = ""
            zip_file_name = f"{''.join(path[:-1])}/{self.name}_code_files.zip"
            zf = zipfile.ZipFile(zip_file_name, mode="w")
            try:
                for file_name in self.local_path_to_code:
                    zf.write(file_name, str(Path(file_name).name))

            except FileNotFoundError:
                logger.error(f"Could not find {FileNotFoundError.filename}")
            finally:
                zf.close()
            with open(zip_file_name, "rb") as f:
                zipped_code = f.read()

            lambda_config = self.lambda_client.update_function_code(
                FunctionName=self.name, ZipFile=zipped_code
            )

        logger.info(f'{lambda_config["FunctionName"]} was updated successfully.')

        return lambda_config

    def _deploy_lambda_to_aws_helper(self, role_res, zipped_code, env_vars):
        """Deploying new lambda to AWS helping function"""
        try:
            lambda_config = self.lambda_client.create_function(
                FunctionName=self.name,
                Runtime=self.runtime,
                Role=role_res["Role"]["Arn"],
                Handler=f"rh_handler_{self.name}.lambda_handler",
                Code={"ZipFile": zipped_code},
                Timeout=self.timeout,
                MemorySize=self.memory_size,
                Environment={"Variables": env_vars},
                EphemeralStorage={"Size": self.tmp_size},  # size of /tmp folder.
            )
            func_name = lambda_config["FunctionName"]
            with self._wait_until_lambda_update_is_finished(func_name):
                lambda_config = self.lambda_client.get_function(FunctionName=func_name)
                return lambda_config
        except self.lambda_client.exceptions.InvalidParameterValueException or botocore.exceptions.ClientError:
            return False

    @contextlib.contextmanager
    def _deploy_lambda_to_aws(self, role_res, zipped_code, env_vars):
        """Deploying new lambda to AWS"""
        config = self._deploy_lambda_to_aws_helper(role_res, zipped_code, env_vars)
        time_passed = 0
        while config is False:
            if time_passed > self.MAX_WAIT_TIME:
                raise TimeoutError(
                    f"Role called {role_res['RoleName']} is being created AWS for too long."
                    + " Please check the resource in AWS console, delete relevant resource(s) "
                    + "if necessary, and re-run your Runhouse code."
                )
            time_passed += 1
            time.sleep(1)
            config = self._deploy_lambda_to_aws_helper(role_res, zipped_code, env_vars)
        yield config

    def _create_role_in_aws(self):
        """create a new role for the lambda function. If already exists - returns it."""

        role_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Action": self.DEFAULT_ROLE_POLICIES,
                    "Resource": "*",
                    "Effect": "Allow",
                }
            ],
        }

        assume_role_policy_document = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {"Service": "lambda.amazonaws.com"},
                    "Action": "sts:AssumeRole",
                }
            ],
        }

        try:
            role_res = self.iam_client.create_role(
                RoleName=f"{self.name}_Role",
                AssumeRolePolicyDocument=json.dumps(assume_role_policy_document),
            )
            logger.info(f'{role_res["Role"]["RoleName"]} was created successfully.')

        except self.iam_client.exceptions.EntityAlreadyExistsException:
            role_res = self.iam_client.get_role(RoleName=f"{self.name}_Role")

        self.iam_client.put_role_policy(
            RoleName=role_res["Role"]["RoleName"],
            PolicyName=f"{self.name}_Policy",
            PolicyDocument=json.dumps(role_policy),
        )

        return role_res

    def _create_new_lambda(self, env_vars):
        """Creates new AWS Lambda."""
        logger.info(f"Creating a new Lambda called {self.name}")
        path = str(Path(self.local_path_to_code[0]).absolute()).split("/")
        path = ["/" + path_e for path_e in path]
        path[0] = ""
        zip_file_name = f"{''.join(path[:-1])}/{self.name}_code_files.zip"
        zf = zipfile.ZipFile(zip_file_name, mode="w")
        try:
            for file_name in self.local_path_to_code:
                zf.write(file_name, str(Path(file_name).name))
        except FileNotFoundError:
            logger.error(f"Could not find {FileNotFoundError.filename}")
        finally:
            zf.close()
        with open(zip_file_name, "rb") as f:
            zipped_code = f.read()

        # creating a role for the Lambda, using default policy.
        # TODO [SB]: in the next phase, enable the user to update the default policy
        role_res = self._create_role_in_aws()

        with self._deploy_lambda_to_aws(
            role_res, zipped_code, env_vars
        ) as lambda_config:
            logger.info(
                f'{lambda_config["Configuration"]["FunctionName"]} was created successfully.'
            )
            Path(zip_file_name).absolute().unlink()
            return lambda_config

    def deploy(self):
        return self.to()

    def to(
        self,
        env: Optional[List[str]] = [],
        cloud: str = "aws_lambda",
        force_install: Optional[bool] = False,
    ):
        """
        Set up a function on AWS as a Lambda function.

        See the args of the factory method :func:`aws_lambda_fn` for more information.
        """

        # Checking if the user have a credentials file
        if not Path(CRED_PATH).is_file():
            logger.error("No credentials found")
            raise FileNotFoundError("No credentials found")
        rh_handler_wrapper = self._rh_wrapper()
        self.local_path_to_code.append(rh_handler_wrapper)

        env_vars = (
            self.env.env_vars if isinstance(self.env, Env) else {"HOME": self.HOME_DIR}
        )

        # if function exist - return its aws config. Else, a new one will be created.
        lambda_config = self._lambda_exist(self.name)
        if not lambda_config:
            # creating a new Lambda function, since it's not existing in the AWS account which is configured locally.
            lambda_config = self._create_new_lambda(env_vars)

        # TODO [SB]: in the next phase, enable the user to change the config of the Lambda.
        # updating the configuration with the initial configuration.
        # lambda_config = self._update_lambda_config(env_vars)

        self.aws_lambda_config = lambda_config
        Path(rh_handler_wrapper).absolute().unlink()
        return self

    # --------------------------------------
    # Lambda Function call methods
    # --------------------------------------

    def __call__(self, *args, **kwargs) -> Any:
        """Call (invoke) the Lambdas function

        Args:
             *args: Optional args for the Function
             **kwargs: Optional kwargs for the Function

        Returns:
            The Function's return value
        """
        return self._invoke(*args, **kwargs)

    def _get_log_group_names(self):
        lambdas_log_groups = self.logs_client.describe_log_groups(
            logGroupNamePrefix=LOG_GROUP_PREFIX
        )["logGroups"]
        lambdas_log_groups = [group["logGroupName"] for group in lambdas_log_groups]
        return lambdas_log_groups

    @contextlib.contextmanager
    def _wait_till_log_group_is_created(
        self, curr_lambda_log_group, lambdas_log_groups
    ):
        time_passed = 0
        while curr_lambda_log_group not in lambdas_log_groups:
            if time_passed > self.MAX_WAIT_TIME:
                raise TimeoutError(
                    f"The log group called {curr_lambda_log_group} is being deployed in AWS cloudwatch for too long. "
                    + " Please check the resource in AWS console, delete relevant resource(s) "
                    + "if necessary, and re-run your Runhouse code."
                )
            time.sleep(1)
            lambdas_log_groups = self._get_log_group_names()
            time_passed += 1
        yield True

    def _invoke(self, *args, **kwargs) -> Any:
        lambdas_log_groups = self._get_log_group_names()
        curr_lambda_log_group = f"{LOG_GROUP_PREFIX}{self.name}"
        invoke_for_the_first_time = curr_lambda_log_group not in lambdas_log_groups

        payload_invoke = {}

        if len(args) > 0 and self.args_names is not None:
            payload_invoke = {self.args_names[i]: args[i] for i in range(len(args))}
        invoke_res = self.lambda_client.invoke(
            FunctionName=self.name,
            Payload=json.dumps({**payload_invoke, **kwargs}),
            LogType="Tail",
        )
        return_value = invoke_res["Payload"].read().decode("utf-8")

        try:
            logger.error(invoke_res["FunctionError"])
            return_value = json.loads(return_value)
            raise RuntimeError(
                f"Failed to run {self.name}. Error: {return_value['errorType']}, {return_value['errorMessage']}"
            )

        except KeyError:
            if invoke_for_the_first_time:
                lambdas_log_groups = self._get_log_group_names()
                with self._wait_till_log_group_is_created(
                    curr_lambda_log_group, lambdas_log_groups
                ):
                    self.logs_client.put_retention_policy(
                        logGroupName=curr_lambda_log_group,
                        retentionInDays=self.retention_time,
                    )

            log_lines = "Function Logs are:\n" + base64.b64decode(
                invoke_res["LogResult"]
            ).decode("utf-8")

            print(log_lines)
            return_value = json.loads(return_value)["body"]
            return return_value

    def map(self, *args, **kwargs):
        """Map a function over a list of arguments.

        Example:
            >>> # The my_lambda_handler.py file
            >>> def my_summer(arg1, arg2, arg3):
            >>>     return arg1 + arg2 + arg3
            >>>
            >>> # your 'main' python file, where you are using runhouse
            >>> summer_lambda = rh.aws_lambda_fn(
            >>>                     fn=my_summer,
            >>>                     runtime = 'python3.9',
            >>>                     name="my_func")
            >>> output = summer_lambda.map([1, 2], [1, 4], [2, 3])  # output = [4, 9]
        """

        return [self._invoke(*args, **kwargs) for args in zip(*args)]

    def starmap(self, args_lists, **kwargs):
        """Like :func:`map` except that the elements of the iterable are expected to be iterables
        that are unpacked as arguments. An iterable of [(1,2), (3, 4)] results in [func(1,2), func(3,4)].

        Example:
            >>> arg_list = [(1,2, 3), (3, 4, 5)]
            >>> # invokes the Lambda function twice, once with args (1, 2, 3) and once with args (3, 4, 5)
            >>> output = summer_lambda.starmap(arg_list) # output = [6, 12]
        """

        return [self._invoke(*args, **kwargs) for args in args_lists]

    # --------------------------------------
    # Lambda Function delete methods
    # --------------------------------------
    def teardown(self):
        """Deletes a Lambda function instance from AWS. All relevant AWS resources
        (role, log group) are deleted as well.

        Example:
            >>> def multiply(a, b):
            >>>     return a * b
            >>> multiply_lambda = rh.aws_lambda_fn(fn=multiply, name="lambdas_mult_func")
            >>> mult_res = multiply_lambda(4, 5)  # returns 20.
            >>> multiply_lambda.teardown()  # returns true if succeeded, raises an exception otherwise.

        """
        try:
            lambda_name = self.name
            # delete from aws console
            if self._lambda_exist(lambda_name):
                policy_name = f"{lambda_name}_Policy"
                role_name = f"{lambda_name}_Role"
                log_group_name = f"/aws/lambda/{lambda_name}"

                # TODO: do we want to delete role and log group?
                #  maybe we want to reuse the role or save logs even if the lambda was deleted.
                self.iam_client.delete_role_policy(
                    RoleName=role_name, PolicyName=policy_name
                )

                self.iam_client.delete_role(RoleName=role_name)

                log_groups = self._get_log_group_names()
                if len(log_groups) > 0 and log_group_name in log_groups:
                    self.logs_client.delete_log_group(logGroupName=log_group_name)

                self.lambda_client.delete_function(FunctionName=lambda_name)

            return True
        except botocore.exceptions.ClientError as aws_exception:
            raise RuntimeError(
                f"Could nor delete an AWS resource, got {aws_exception.response['Error']['Message']}"
            )

    def config(self, condensed=True):
        config = super().config(condensed)
        config.update(
            {
                "paths_to_code": self.local_path_to_code,
                "handler_function_name": self.handler_function_name,
                "runtime": self.runtime,
                "timeout": self.timeout,
                "memory_size": self.memory_size,
                "tmp_size": self.tmp_size,
                "args_names": self.args_names,
                "function_arn": self.aws_lambda_config["Configuration"]["FunctionArn"],
            }
        )
        return config

    # --------------------------------------
    # Properties setup
    # --------------------------------------

    @property
    def handler_path(self):
        return self.local_path_to_code[0]
